/*
                            XSW Object Matching

	Functions:

	int DBIsObjectGarbage(long object_num)
	double DBGetObjectVisibility(long object_num)
	double DBGetObjectVisibilityPtr(xsw_object_struct *obj_ptr)

	long MatchWeaponsLock(
		long ref_obj,
		long start_obj,
		double range
	)
	long MatchIntercept(long object_num, char *arg)
	long MatchInterceptByNumber(long ref_obj, long tar_obj)
	long MatchObjectByName(char *name, int type)

	void DBSetupGarbageObject()

	---

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/reality.h"
#include "../include/mathutil.h"
#include "../include/isrefs.h"
#include "../include/dbmatch.h"




/*
 *      Checks if object is allocated and garbage.
 *
 *      Returns true if not allocated or garbage.
 */
int DBIsObjectGarbage(long object_num)
{
        if((object_num < 0) ||
           (object_num >= total_objects) ||
           (xsw_object == NULL)
        )
            return(-1);
        else if(xsw_object[object_num] == NULL)
            return(-1);
        else if(xsw_object[object_num]->type <= XSW_OBJ_TYPE_GARBAGE)
            return(1);
        else
            return(0);
}


/*           
 *      Returns the object's visibility.
 */         
double DBGetObjectVisibility(long object_num)
{
        if(DBIsObjectGarbage(object_num))
            return(0.00);
        else 
            return(DBGetObjectVisibilityPtr(xsw_object[object_num]));
}

double DBGetObjectVisibilityPtr(xsw_object_struct *obj_ptr)
{
        if(obj_ptr == NULL)
            return(0.00);

        if(obj_ptr->type == XSW_OBJ_TYPE_AREA)
            return(obj_ptr->cur_visibility);
	else if(obj_ptr->loc_type == XSW_LOC_TYPE_NEBULA)
	    return(obj_ptr->cur_visibility * VISIBILITY_NEBULA);
        else
            return(obj_ptr->cur_visibility);
}


/*
 *      returns the next XSW Object number that is within
 *      range range of object ref_object.   ref_object should be the object
 *      that is doing the weapons lock, and may never be -1.
 *      start_object maybe -1 or any object number less than total_objects.
 *      range should be the the maximum scanner range of ref_object.
 *      Returns the next XSW Object number of a valid (nongarbage or nonerror)
 *      object withing range range of ref_object starting from start_object,
 *      or returns -1 if no match can be made.
 */
long MatchWeaponsLock(long ref_obj, long start_obj, double range) 
{
	long i, f;
	xsw_object_struct *obj_ptr, *ref_obj_ptr;


	if(DBIsObjectGarbage(ref_obj))
	    return(-1);
	else
	    ref_obj_ptr = xsw_object[ref_obj];

	/* Sanitize start object (can be -1). */
        if(start_obj < -1)
            start_obj = -1;

	/* range must be positive. */
	if(range < 0)
	    range = 1;


	/* Begin search, note that start_obj could be -1. */
	for(i = start_obj + 1, f = -1;
            i < total_objects;
            i++
	)
	{
	    obj_ptr = xsw_object[i];

	    if(obj_ptr == NULL)
		continue;
	    if(obj_ptr->type <= XSW_OBJ_TYPE_GARBAGE)
		continue;

	    if((obj_ptr->type == XSW_OBJ_TYPE_WEAPON) ||
               (obj_ptr->type == XSW_OBJ_TYPE_STREAMWEAPON) ||
               (obj_ptr->type == XSW_OBJ_TYPE_SPHEREWEAPON) ||
               (obj_ptr->type == XSW_OBJ_TYPE_AREA)
	    )
		continue;

	    /* Don't match referance object. */
	    if(obj_ptr == ref_obj_ptr)
		continue;

	    /* See if object is in range. */
	    if(Mu3DInRangePtr(obj_ptr, ref_obj_ptr,
		    range * DBGetObjectVisibilityPtr(obj_ptr)
	       )
	    )
	    {
	        /* Checks passed, this is the object we want. */
	        f = i;
	        break;
	    }
	}


	return(f);
}



/*
 *	Tries to match object specified in arg.  If arg
 *	is "" or "#off" then -1 is returned.   If the object specified in arg
 *	is found and is of type XSW_OBJ_TYPE_STATIC then that object's number
 *	is returned.  If the object is found but not of type static, then it 
 *	is checked to see if it is within scanner range of object_num and if 
 *	it is, then it's number is returned or of not in range then -1 is    
 *	returned.
 */
long MatchIntercept(long object_num, char *arg)
{
	char name[XSW_OBJ_NAME_MAX];
	long target_obj;
	xsw_object_struct *obj_ptr;
	xsw_object_struct *tar_obj_ptr;


	if(arg == NULL)
	    return(-1);

	if(DBIsObjectGarbage(object_num))
            return(-1);
	else
	    obj_ptr = xsw_object[object_num];


	/* Sanitize copy over name to be searched. */
	strncpy(name, arg, XSW_OBJ_NAME_MAX);
	name[XSW_OBJ_NAME_MAX - 1] = '\0';
	StringStripSpaces(name);


	/* Turn intercept off? */
	if(!strcasecmp(name, "#off") ||
	   (name[0] == '\0')
	)
	    return(-1);


	/* Match target_obj. */
	target_obj = MatchObjectByName(name, -1);
	if(DBIsObjectGarbage(target_obj))
	    return(-1);
	else
            tar_obj_ptr = xsw_object[target_obj];


	/* Cannot intercept itself. */
	if(target_obj == object_num)
	    return(-1);


	/* Following types of objects may not be intercepted. */
	if((tar_obj_ptr->type == XSW_OBJ_TYPE_WEAPON) ||
	   (tar_obj_ptr->type == XSW_OBJ_TYPE_STREAMWEAPON) ||
           (tar_obj_ptr->type == XSW_OBJ_TYPE_SPHEREWEAPON)
	)
	    return(-1);

        /* Following objects don't need range check. */
        if((tar_obj_ptr->type == XSW_OBJ_TYPE_STATIC) ||
	   (tar_obj_ptr->type == XSW_OBJ_TYPE_HOME) ||
           (tar_obj_ptr->type == XSW_OBJ_TYPE_AREA)
	)
            return(target_obj);


        /* Is target_obj within scanner range of object_num? */
        if(Mu3DInRangePtr(obj_ptr, tar_obj_ptr,
                obj_ptr->scanner_range *
                DBGetObjectVisibilityPtr(tar_obj_ptr)
            )
        )
	    return(target_obj);
	else
	    return(-1);
}


/*
 *	Checks if tar_obj is within intercept range of ref_obj,
 *	and that both are valid.  Returns tar_obj if everything
 *	checks out okay or -1 if not.
 */
long MatchInterceptByNumber(long ref_obj, long tar_obj)
{
        xsw_object_struct *obj_ptr;
        xsw_object_struct *tar_obj_ptr;


        /* Error checks. */
        if(DBIsObjectGarbage(ref_obj))
            return(-1);
        else
            obj_ptr = xsw_object[ref_obj];

        if(DBIsObjectGarbage(tar_obj))
            return(-1);
        else
            tar_obj_ptr = xsw_object[tar_obj];


        /* ref_obj and tar_obj cannot be the same. */
        if(ref_obj == tar_obj)
            return(-1);


        /* Skip these types of objects. */
        if((tar_obj_ptr->type == XSW_OBJ_TYPE_WEAPON) ||
           (tar_obj_ptr->type == XSW_OBJ_TYPE_STREAMWEAPON) ||
           (tar_obj_ptr->type == XSW_OBJ_TYPE_SPHEREWEAPON)
        )
            return(-1);


        /* Following objects always matchable regardless of distance. */
        if((tar_obj_ptr->type == XSW_OBJ_TYPE_STATIC) ||
	   (tar_obj_ptr->type == XSW_OBJ_TYPE_HOME) ||
	   (tar_obj_ptr->type == XSW_OBJ_TYPE_AREA)
	)
            return(tar_obj);


        /* If none of the above, then do scanner range check. */
        if(Mu3DInRangePtr(obj_ptr, tar_obj_ptr,
                obj_ptr->scanner_range *
                DBGetObjectVisibilityPtr(tar_obj_ptr)
            )
        )
	    return(tar_obj);
	else
            return(-1);
}


/*
 *	Matches object by name, return its number or -1 on error or
 *	no match.
 *
 *	The search can be narrowed down to a certain type of object
 *	or -1 for all types.
 */
long MatchObjectByName(char *name, int type)
{
	long object_num;
	xsw_object_struct **obj_ptr;


	/* Name cannot be NULL or contain no characters. */
	if(name == NULL)
	    return(-1);
	if(name[0] == '\0')
	    return(-1);


	/* Skip leading spaces in name. */
	while(isblank(*name))
	    name++;

	/* Number matching? */
	if((*name == '#') ||
           (isdigit(*name))
	)
	{
	    /* Skip leading '#' characters. */
	    while(*name == '#')
		name++;

	    /* Get object number. */
	    object_num = atol(name);

	    /* Check if object_num is valid. */
	    if((object_num < 0) || (object_num >= total_objects))
		return(-1);
	    if(xsw_object[object_num] == NULL)
		return(-1);
	    if(xsw_object[object_num]->type <= XSW_OBJ_TYPE_GARBAGE)
		return(-1);

            /* Match all types of objects? */
	    if(type == -1)
		return(object_num);

	    /* Match specific type. */
	    if(xsw_object[object_num]->type == type)
		return(object_num);
	    else
		return(-1);
	}


	/* Begin searching by name. */
	for(object_num = 0, obj_ptr = xsw_object;
            object_num < total_objects;
            object_num++, obj_ptr++
	)
	{
	    if(*obj_ptr == NULL) continue;
	    if((*obj_ptr)->type <= XSW_OBJ_TYPE_GARBAGE) continue;

	    /* Object must characters in its name. */
	    if((*obj_ptr)->name[0] == '\0') continue;

	    /* Match all types? */
	    if(type == -1)
	    {
		/* Partial name match. */
		if(strstr((*obj_ptr)->name, name) == NULL)
		    continue;

		/* Got match. */
		return(object_num);
	    }
	    /* Match specific type. */
	    else
	    {
		/* Match type first. */
		if((*obj_ptr)->type != type)
		    continue;

                /* Partial name match. */
                if(strstr((*obj_ptr)->name, name) == NULL)
                    continue;

                /* Got match. */
                return(object_num);
	    }
	}


	/* No match. */
	return(-1);
}


/*
 *	Sets up the global garbage object.
 */
void DBSetupGarbageObject()
{
        bzero(garbage_object.name, XSW_OBJ_NAME_MAX);
        bzero(garbage_object.empire, XSW_OBJ_PASSWORD_MAX); 
	bzero(garbage_object.password, XSW_OBJ_EMPIRE_MAX);

        garbage_object.last_updated = 0;

        garbage_object.type = XSW_OBJ_TYPE_GARBAGE;
	garbage_object.loc_type = XSW_LOC_TYPE_SPACE;
        garbage_object.imageset = ISREF_DEFAULT;
        garbage_object.owner = -1;
        garbage_object.size = 1;
        garbage_object.locked_on = -1;  
        garbage_object.intercepting_object = -1;

        garbage_object.scanner_range = 8.0;
        garbage_object.sect_x = 0;
        garbage_object.sect_y = 0;
        garbage_object.sect_z = 0;
        garbage_object.x = 0;
        garbage_object.y = 0;
        garbage_object.z = 0;
        garbage_object.heading = 0;
        garbage_object.pitch = 0;
        garbage_object.bank = 0;
        garbage_object.velocity = 0;
        garbage_object.velocity_max = 0;
        garbage_object.velocity_heading = 0;
        garbage_object.thrust_dir = 3.1415927;
        garbage_object.thrust = 0;   
        garbage_object.thrust_power = 0;   
        garbage_object.thrust_power = 0;
        garbage_object.throttle = 0;
	garbage_object.engine_state = ENGINE_STATE_ON;
        garbage_object.turnrate = 0.0003;
        garbage_object.hp = 1;
        garbage_object.hp_max = 1;
        garbage_object.power = 0;
        garbage_object.power_max = 0;
        garbage_object.power_purity = 1.0;
        garbage_object.core_efficency = 0;
        garbage_object.antimatter = 0;
        garbage_object.antimatter_max = 0;
        garbage_object.shield_state = SHIELD_STATE_NONE;
        garbage_object.shield_frequency = 180.25;
        garbage_object.selected_weapon = 0;
        garbage_object.total_weapons = 0;
        garbage_object.birth_time = 0;
        garbage_object.lifespan = -1;
        garbage_object.cloak_state = CLOAK_STATE_NONE;   
        garbage_object.cloak_strength = 0.0;
        garbage_object.shield_visibility = 0.0;
        garbage_object.visibility = 1.0;
	garbage_object.cur_visibility = 1.0;
        garbage_object.damage_control = DMGCTL_STATE_OFF;
	garbage_object.com_channel = COM_CHANNEL_MIN;

	garbage_object.tractored_object = NULL;
	garbage_object.total_tractored_objects = 0;

        garbage_object.permission.uid = DEFAULT_UID;
        garbage_object.permission.gid = DEFAULT_GID;

        garbage_object.animation.interval = 500;
        garbage_object.animation.last_interval = 0;
        garbage_object.animation.current_frame = 0;
        garbage_object.animation.total_frames = 1;
        garbage_object.animation.cycle_count = 0;
        garbage_object.animation.cycle_times = -1;	/* Infinatly. */

        garbage_object.weapons = NULL;
	garbage_object.score = NULL;
	garbage_object.eco = NULL;


	return;
}
