/*
                       Game Controller Management

	Functions:

	int GCtlInit(int controller)
	void GCtlUpdate()
	void GCtlShutdown()

	---

	These functions serve as a front end for other functions
	to easilly tell what controller state (not needing to know
	which type of controller is in use) the currect controller
	is in by checking the values of the members of the gctl
	structures.

 */

#include "keymap.h"

#include "xsw.h"


/*
 *	Null zone coefficients (in units OF coefficent):
 */
static double x_nz_coeff, y_nz_coeff;



/*
 *	Switches to and initializes controller.
 *	Previous controller will be automatically shut down.
 *
 *	Returns non zero on error.
 */
int GCtlInit(int controller)
{
	int status, prev_controller;


	/* Record previous controller type. */
	prev_controller = option.controller;


	/* Initialize new controller. */
	switch(controller)
	{
#ifdef JS_SUPPORT
	  /* *********************************************************** */
	  case CONTROLLER_JOYSTICK:
	    /* Shutdown pointer and keyboard. */

	    /* Shutdown and reinitialize joystick. */
            JSClose(&joystick_data[0]);
	    status = JSInitialize(
                DEV_JOYSTICK1,		/* Device node. */
                fname.js_calib,	/* Caliberation file. */
                &joystick_data[0]	/* Joystick data. */
	    );
	    if(joystick_data[0].fd < 0)
		return(-1);

	    /*   Calculate local x and y null zone coefficient values.
	     *   Note that joystick_data[0].nullzone is the radius.
	     */
	    if(joystick_data[0].x_scope != 0)
	        x_nz_coeff = (double)joystick_data[0].nullzone /
		    (double)joystick_data[0].x_scope;
	    else
		x_nz_coeff = 0;

            if(joystick_data[0].y_scope != 0)
                y_nz_coeff = (double)joystick_data[0].nullzone /
                    (double)joystick_data[0].y_scope;
            else
                y_nz_coeff = 0;

	    break;
#endif /* JS_SUPPORT */

	  /* *********************************************************** */
	  case CONTROLLER_POINTER:
	    /* Shutdown joystick and keyboard. */

	    /* Not supported yet. */
	    return(-2);

	    break;

	  /* *********************************************************** */
	  default: /* Default to CONTROLLER_KEYBOARD. */
	    /* Shutdown joystick and pointer. */
#ifdef JS_SUPPORT
	    JSClose(&joystick_data[0]);
#endif /* JS_SUPPORT */

	    /* Initialize keyboard. */

	    break;
	}


	/*   Need to update new controller being used in global
	 *   option.controller
	 */
	option.controller = controller;


	/* Reset all gctls. */
	bzero(&gctl[0], sizeof(gctl_struct));
        bzero(&gctl[1], sizeof(gctl_struct));
        bzero(&gctl[2], sizeof(gctl_struct));
        bzero(&gctl[3], sizeof(gctl_struct));


	return(0);
}



/*
 *	Gets controller state and updates all gctl structures.
 *
 *	Global settings such as keymappings and throttle settings
 *	are taken into account and calculated!
 *
 */
void GCtlUpdate()
{
	double tmp_x_coeff, tmp_y_coeff;

	char button1, button2, button3, button4;

	/* Previous key states. */
	static char	pks_turn_left,
			pks_turn_right,
			pks_throttle_inc,
			pks_throttle_dec,
			pks_fire_weapon,
			pks_omni_dir_thrust,
			pks_external_dampers;

	double	pks_x_accel, pks_y_accel;

	/* For keyboard event handling from GUI. */
	event_t event;
	char key_state;		/* 0 = released, 1 = pressed. */


	/* Must reset all button activation records. */
	gctl[0].button1_act = 0;
	gctl[0].button2_act = 0;
	gctl[0].button3_act = 0;
	gctl[0].button4_act = 0;
        gctl[0].button5_act = 0;
        gctl[0].button6_act = 0;
        gctl[0].button7_act = 0;
        gctl[0].button8_act = 0;

/* In the future, add reset for all gctls. */


	switch(option.controller)
	{
#ifdef JS_SUPPORT
          /* ************************************************************** */
          case CONTROLLER_JOYSTICK:
	    if(joystick_data[0].fd < 0)
		break;

	    /* Get x and y positions. */
	    JSGetPosition(
		&joystick_data[0],
		&gctl[0].x,
		&gctl[0].y,
		&button1,
		&button2,
		&button3,
		&button4
	    );

	    /* Handle null zone for x. */
	    tmp_x_coeff = gctl[0].x;
	    if((tmp_x_coeff >= (x_nz_coeff * -1)) &&
               (tmp_x_coeff <= (x_nz_coeff * 1))
	    )
	    {
		/* X axis is in null zone. */
		gctl[0].x = 0;
	    }
	    else
	    {
		/* X axis is out of null zone. */
		if(tmp_x_coeff < 0)
		{
		    /* Left. */
                    if((1 - x_nz_coeff) > 0)
                        gctl[0].x = (tmp_x_coeff + x_nz_coeff) / (1 - x_nz_coeff);
                    else
                        gctl[0].x = 0;
		}
		else
		{
		    /* Right. */
		    if((1 - x_nz_coeff) > 0)
			gctl[0].x = (tmp_x_coeff - x_nz_coeff) / (1 - x_nz_coeff);
		    else
			gctl[0].x = 0;
		}
	    }


            /* Handle NULL zone for y (only when in bi-directional throttle). */
	    if(option.throttle_mode == THROTTLE_MODE_BIDIRECTIONAL)
	    {
		tmp_y_coeff = gctl[0].y;
                if((tmp_y_coeff >= (y_nz_coeff * -1)) &&
                   (tmp_y_coeff <= (y_nz_coeff * 1))
                )
		{
                    /* Y axis is in null zone. */
                    gctl[0].y = 0;
		}
		else
		{
                    /* Y axis is out of null zone. */
                    if(tmp_y_coeff < 0)
                    {
                        /* Down. */
                        if((1 - y_nz_coeff) > 0)
                            gctl[0].y = (tmp_y_coeff + y_nz_coeff) / (1 - y_nz_coeff);
                        else
                            gctl[0].y = 0;
                    }
                    else
                    {
                        /* Up. */
                        if((1 - y_nz_coeff) > 0)
                            gctl[0].y = (tmp_y_coeff - y_nz_coeff) / (1 - y_nz_coeff);
                        else
                            gctl[0].y = 0;
                    }
		}
	    }
	    /* Adjust y value for THROTTLE_MODE_NORMAL. */
	    else
	    {
		/* Adjust range of gctl[0].y to 0.0 to 1.0. */
		gctl[0].y = (double)(gctl[0].y + 1) / (double)2;
	    }

	    /* Set standard joystick button states. */
	    if(button1)
	    {
		if(!gctl[0].button1)
                    gctl[0].button1_act = 1;
		gctl[0].button1 = 1;
	    }
	    else
            {
                gctl[0].button1 = 0;
            }

            if(button2)
            {
               if(!gctl[0].button2)
                    gctl[0].button2_act = 1;
                gctl[0].button2 = 1;
            }
            else
            {
                gctl[0].button2 = 0;
            }

            if(button3)
            {
               if(!gctl[0].button3)
                    gctl[0].button3_act = 1;
                gctl[0].button3 = 1;
            }
            else
            {
                gctl[0].button3 = 0;
            }

            if(button4)
            {
               if(!gctl[0].button4)
                    gctl[0].button4_act = 1;
                gctl[0].button4 = 1;
            }
            else
            {
                gctl[0].button4 = 0;
            }

            break;
#endif /* JS_SUPPORT */

          /* ************************************************************** */
          case CONTROLLER_POINTER:
            /* Not supported yet. */
            break;

	  /* ************************************************************** */
          default: /* Default to CONTROLLER_KEYBOARD. */
	    /* Break if GUI is not connected. */
	    if(!IDC())
		break;

	    while(OSWEventsPending() > 0)
	    {
		/* Check and get Key events only, break if no more exist. */
	        if(!OSWCheckMaskEvent(
		    KeyPressMask | KeyReleaseMask,
		    &event
		   )
		)
		     break;

		/* Get key state. */
		key_state = ((event.type == KeyPress) ? 1 : 0);


                /* Turn left. */
                if(event.xkey.keycode == xsw_keymap[XSW_KM_TURN_LEFT].keycode)
                {
                    if(key_state)
                    {
			OSWKBAutoRepeatOff();
		        pks_turn_left = 1;
                    }
                    else
                    {
		        pks_turn_left = 0;
                    }
                }
                /* Turn right. */
                else if(event.xkey.keycode == xsw_keymap[XSW_KM_TURN_RIGHT].keycode)
                {
		    if(key_state)
		    {
                        OSWKBAutoRepeatOff();
                        pks_turn_right = 1;
		    }
		    else
                    {
			pks_turn_right = 0;
                    }
                }
	        /* Throttle up. */
                else if(event.xkey.keycode == xsw_keymap[XSW_KM_THROTTLE_INC].keycode)
                {
                    if(key_state)
                    {
                        OSWKBAutoRepeatOff();
		        pks_throttle_inc = 1;
                    }
                    else
                    {
		        pks_throttle_inc = 0;
                    }
                }
                /* Throttle down. */
                else if(event.xkey.keycode == xsw_keymap[XSW_KM_THROTTLE_DEC].keycode)
                {
                    if(key_state)
                    {
                        OSWKBAutoRepeatOff();
                        pks_throttle_dec = 1;
                    }
                    else
                    {
                        pks_throttle_dec = 0;
                    }
                }
                /* Fire weapon (button 1). */
                else if(event.xkey.keycode == xsw_keymap[XSW_KM_FIRE_WEAPON].keycode)
                {
                    if(key_state)
                    {
                       if(!gctl[0].button1)
                            gctl[0].button1_act = 1;
                        gctl[0].button1 = 1;
                        OSWKBAutoRepeatOff();
			pks_fire_weapon = 1;
                    }
                    else
                    {
                        gctl[0].button1 = 0;
                        pks_fire_weapon = 0;
                    }
                }
                /* Omni directrional thrust (button 2). */
		else if(event.xkey.keycode == xsw_keymap[XSW_KM_OMNI_DIR_THRUST].keycode)
		{
                    if(key_state)
                    {
			if(!gctl[0].button2)
			    gctl[0].button2_act = 1;
                        gctl[0].button2 = 1;
                        OSWKBAutoRepeatOff();
                        pks_omni_dir_thrust = 1;
                    }
                    else
                    {
                        gctl[0].button2 = 0;
                        pks_omni_dir_thrust = 0;
                    }
		}
                /* External dampers (button 3). */
                else if(event.xkey.keycode == xsw_keymap[XSW_KM_EXTERNAL_DAMPERS].keycode)
                {
                    if(key_state)
                    {
                        if(!gctl[0].button3)
                            gctl[0].button3_act = 1;
                        gctl[0].button3 = 1;
                        OSWKBAutoRepeatOff();
                        pks_external_dampers = 1;
                    }
                    else
                    {
                        gctl[0].button3 = 0;
                        pks_external_dampers = 0;
                    }
                }
		else
		{
		    OSWPutBackEvent(&event);
		    break;
		}
	    }

	    /* No button4 for keyboard. */
            gctl[0].button4 = 0;

	    /* Turn auto repeat back on only when all states false. */
	    if(!pks_turn_left &&
               !pks_turn_right &&
               !pks_throttle_inc &&
               !pks_throttle_dec &&
               !pks_fire_weapon &&
               !pks_omni_dir_thrust &&
               !pks_external_dampers
	    )
		OSWKBAutoRepeatOn();


	    /* ********************************************************* */
	    /* Actually update game controller states (except buttons). */

	    pks_x_accel = gctl[0].x;

	    /* Turn left (x). */
	    if(pks_turn_left)
	    {
		pks_x_accel -= 0.02 * time_compensation;
	    }
	    /* Turn right (x). */
	    else if(pks_turn_right)
	    {
		pks_x_accel += 0.02 * time_compensation;
	    }
	    /* Turn left (x) decelleration. */
	    else if(pks_x_accel < 0)
	    {
		pks_x_accel += 0.02 * time_compensation;
	    }
	    /* Turn right (x) decelleration. */
            else if(pks_x_accel > 0)
            {
                pks_x_accel -= 0.02 * time_compensation;
            }

	    /* Sanitize accelerations (x). */
	    if(pks_x_accel > 1)
		pks_x_accel = 1;
	    if(pks_x_accel < -1)
		pks_x_accel = -1;
	    if((pks_x_accel > (-0.01 * time_compensation)) &&
               (pks_x_accel < (0.01 * time_compensation))
	    )
		pks_x_accel = 0;

	    gctl[0].x = pks_x_accel;



            pks_y_accel = gctl[0].y;

	    /* Turottle increase (y). */
            if(pks_throttle_inc)
            {
                pks_y_accel += ((osw_gui[0].shift_key_state) ? 0.01 : 0.002)
		    * time_compensation;
                if(pks_y_accel > 1)
                    pks_y_accel = 1;
            }
	    /* Throttle decrease (y). */
            if(pks_throttle_dec)
            {
                pks_y_accel -= ((osw_gui[0].shift_key_state) ? 0.01 : 0.002)
		    * time_compensation;
		if(pks_y_accel < -1)
		    pks_y_accel = -1;
            }
	    /* Throttle null zone (y), if was not 0. */
	    if((pks_y_accel < 0.005) &&
               (pks_y_accel > -0.005) &&
               (gctl[0].y != 0)
	    )
		pks_y_accel = 0;

	    gctl[0].y = pks_y_accel;


            break;
	}



	return;
}



/*
 *	Closes all game controllers.
 */
void GCtlShutdown()
{


#ifdef JS_SUPPORT
	/* Close joystick. */
	JSClose(&joystick_data[0]);
#endif /* JS_SUPPORT */

	/* Close pointer device. */

	/* Set global options controller setting to something sane. */
	option.controller = CONTROLLER_KEYBOARD;


	return;
}
