/*
                            Joystick Low-level

	Functions:

	int JSInitialize(
		char *devname,
		char *calfile,
		joystick_data_struct *jsd
	)
	void JSGetPosition(
	        joystick_data_struct *jsd,
	        double *x,
	        double *y,
	        char *button1,
	        char *button2,
	        char *button3,
	        char *button4
	)
	void JSClose(joystick_data_struct *jsd)

	---



 */

#ifdef JS_SUPPORT


#include "xsw.h"


/* Joystick header file. */ 
#if defined(__linux__)
#  include <linux/joystick.h>
#endif
#if defined(__HPUX__)

#endif

#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))


/*
 *	Initializes the joystick at the device node jsdevicenode.
 *	Uses the caliberation file calfile and fills the
 *	joystick_data_struct jss with the joystick information.
 */
#if defined(__linux__)
int JSInitialize(
	char *devname,
	char *calfile,
	joystick_data_struct *jsd
)
{
	char *strptr, *strptr2;
	FILE *fp;
        off_t filesize;
        struct stat stat_buf;

	char parm[CFG_PARAMETER_MAX];
	char val[CFG_VALUE_MAX];

        int jsfd = -1;
	int lines_read = 0;


	/* Error checks. */
	if((devname == NULL) ||
	   (calfile == NULL) ||
           (jsd == NULL)
	)
	    return(-1);


	/* *********************************************************** */

        /* Reset all joystick data values. */
        jsd->fd = -1;
        jsd->x_min = 0;
        jsd->x_max = 1200;
        jsd->y_min = 0;
        jsd->y_max = 1200;
        jsd->nullzone = 200;
        jsd->button1_state = JS_FALSE;
        jsd->button2_state = JS_FALSE;
        jsd->button3_state = JS_FALSE;
        jsd->button4_state = JS_FALSE;
        jsd->x_trim = 0;
        jsd->y_trim = 0;


	/* Does devname exist?. */
        if(stat(devname, &stat_buf))
        {
            fprintf(stderr,
                "%s: No such device.\n",
                devname
            );
            return(-1);
	}

	/* Does calfile exist? */
        if(stat(calfile, &stat_buf))
        {
	    fprintf(stderr,
                "%s: No such file.\n",
		calfile
	    );
            return(-1);
        }

	/* Get size of the caliberation file. */
        filesize = stat_buf.st_size;

	/* Open caliberation file. */
	fp = fopen(calfile, "rb");
        if(fp == NULL)
        {
            fprintf(stderr,
               "%s: Cannot open.\n",
               calfile
            );
            return(-1);   
        }


	/* ********************************************************* */

	strptr = NULL;

        while(1)
        {
            /* Free previous line and allocate/read next line. */
            free(strptr); strptr = NULL;
            strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
            if(strptr == NULL) break;
            lines_read++;
            
            /* Fetch parameter. */
            strptr2 = StringCfgParseParm(strptr);
            if(strptr2 == NULL) continue;
            strncpy(parm, strptr2, CFG_PARAMETER_MAX);
            parm[CFG_PARAMETER_MAX - 1] = '\0';
                 
            /* Fetch value. */
            strptr2 = StringCfgParseValue(strptr);
            if(strptr2 == NULL) strptr2 = "0";  /* Set it to "0" if NULL. */
            strncpy(val, strptr2, CFG_VALUE_MAX);
            val[CFG_VALUE_MAX - 1] = '\0';


	    /* BeginJoystick1 */
	    if(!strcmp(parm, "BeginJoystick1"))
	    {
                while(1)
                {
                    /* Free previous line and allocate/read next line. */
                    free(strptr); strptr = NULL;
                    strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
                    if(strptr == NULL) break;
                    lines_read++;

                    /* Fetch parameter. */
                    strptr2 = StringCfgParseParm(strptr);
                    if(strptr2 == NULL) continue;
                    strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                    parm[CFG_PARAMETER_MAX - 1] = '\0';

                    /* Fetch value. */
                    strptr2 = StringCfgParseValue(strptr);
                    if(strptr2 == NULL) strptr2 = "0";
                    strncpy(val, strptr2, CFG_VALUE_MAX);
                    val[CFG_VALUE_MAX - 1] = '\0';


		    /* X_Min */
	       	    if(!strcmp(parm, "X_Min"))
	            {
			jsd->x_min = atoi(val);
           	    }
                    /* X_Center */
                    else if(!strcmp(parm, "X_Center"))
                    {
                        jsd->x_center = atoi(val);
                    }
                    /* X_Max */
                    else if(!strcmp(parm, "X_Max"))
                    {
                        jsd->x_max = atoi(val);
                    }

	            /* Y_Min */
        	    else if(!strcmp(parm, "Y_Min"))
	            {
			jsd->y_min = atoi(val);
                    }
                    /* Y_Center */
                    else if(!strcmp(parm, "Y_Center"))
                    {
			jsd->y_center = atoi(val);
                    }
                    /* Y_Max */
                    else if(!strcmp(parm, "Y_Max"))
                    {
			jsd->y_max = atoi(val);
                    }

                    /* NullZone */
                    else if(!strcmp(parm, "NullZone"))
                    {
			jsd->nullzone = atoi(val);
                    }

                    /* EndJoystick1 */
                    else if(!strcmp(parm, "EndJoystick1"))
		    {
			/* End of this joystick entry. */
			break;
                    }

		    /* Unknown parameter. */
		    else
		    {
			/* Do not warn. */
		    }
                }
	    }
	}

	/* Close the file. */
        fclose(fp);
	fp = NULL;


	/* ********************************************************* */
	/* Sanitize new joystick data parameters. */

	/* Minimums cannot be greater than maximums. */
	if(jsd->x_min > jsd->x_max)
	{
	    fprintf(stderr,
 "%s: Warning: X minumum %i is greater than X maximum %i.\n",
		calfile,
		jsd->x_min,
		jsd->x_max
	    );

	    /* Set maximum to the minimum's value. */
	    jsd->x_max = jsd->x_min;
	}
        if(jsd->y_min > jsd->y_max)
        {
            fprintf(stderr,
 "%s: Warning: Y minimum is greater than Y maximum %i.\n",
                calfile,
                jsd->y_max
            );

            /* Set maximum to the minimum's value. */
            jsd->y_max = jsd->y_min;
        }

	/* Calculate X and Y scopes. */
	jsd->x_scope = jsd->x_max - jsd->x_min;
        jsd->y_scope = jsd->y_max - jsd->y_min;   

  

	/* ********************************************************** */
	/* Open the joystick device. */

	jsfd = open(devname, O_RDONLY);
	if(jsfd < 0)
	{
	    fprintf(stderr,
		"%s: Cannot open device.\n",
		devname
	    );
	    return(-1);
 	}
	else
	{
            jsd->fd = jsfd;
	}


	/* Return the joystick file descriptor. */
	return(jsfd);

}
#endif /* __linux__ */



#if defined(__linux__)
/*
 *	Gets joystick coordinate and button states.
 *
 *	Coordinates x and y are returned in the range -1.0 to 1.0
 *	using left hand rule.
 */
void JSGetPosition(
	joystick_data_struct *jsd,
	double *x,
	double *y,
	char *button1,
	char *button2,
	char *button3,
	char *button4
)
{
	struct JS_DATA_TYPE js;


	if(jsd == NULL)
	{
	    if(x != NULL)
		*x = 0;
	    if(y != NULL)
		*y = 0;
	    if(button1 != NULL)
		*button1 = 0;
            if(button2 != NULL)
                *button2 = 0;
            if(button3 != NULL)
                *button3 = 0;
            if(button4 != NULL)
                *button4 = 0;

	    return;
	}

	/* Read joystick positions. */
        if(read(jsd->fd, &js, JS_RETURN) < 0)
        {
            if(x != NULL)
                *x = 0;
            if(y != NULL)
                *y = 0;
            if(button1 != NULL)
                *button1 = 0;
            if(button2 != NULL)
                *button2 = 0;
            if(button3 != NULL)
                *button3 = 0;
            if(button4 != NULL)
                *button4 = 0;
            
            return;
        }


	/* Calculate x coordinate. */
	if(x != NULL)
	{
	    js.x += jsd->x_trim;	/* Apply trim. */

            if(js.x < jsd->x_center)
            {
                *x = (double)((double)(js.x - jsd->x_min) /
                    (double)MAX(jsd->x_center - jsd->x_min, 1)) - 1;
            }
            else
            {
                *x = (double)(js.x - jsd->x_center) /
                    (double)MAX(jsd->x_max - jsd->x_center, 1);
            }

	    /* Sanitize. */
	    if(*x > 1) *x = 1;
	    if(*x < -1) *x = -1;
	}

        /* Calculate y coordinate. */
        if(y != NULL)
        {
            js.y += jsd->y_trim;	/* Apply trim. */

            if(js.y < jsd->y_center)
            {
                *y = 1 - (double)((double)(js.y - jsd->y_min) /
                    (double)MAX(jsd->y_center - jsd->y_min, 1));
            }
            else
            {
                *y = (double)(js.y - jsd->y_center) /
                    (double)MAX(jsd->y_max - jsd->y_center, 1) * -1;
            }

            /* Sanitize. */
            if(*y > 1) *y = 1;
            if(*y < -1) *y = -1;
        }


	/* Button states. */
	if(button1 != NULL)
	    *button1 = (js.buttons & (1 << 0)) ? 1 : 0;

        if(button2 != NULL)
            *button2 = (js.buttons & (1 << 1)) ? 1 : 0;

        if(button3 != NULL)
            *button3 = (js.buttons & (1 << 2)) ? 1 : 0;

        if(button4 != NULL)
            *button4 = (js.buttons & (1 << 3)) ? 1 : 0;


	return;
}
#endif /* __linux__ */


/*
 *	Closes the joystick device specified with the descriptor
 *	int jsd.
 */
#if defined(__linux__)
void JSClose(joystick_data_struct *jsd)
{

	if(jsd == NULL)
	    return;
	if(jsd->fd < 0)
	    return;

	/* Close joystick device. */
	close(jsd->fd);
	/* Mark joystick as uninitialized. */
	jsd->fd = -1;

	return;
}
#endif /* __linux__ */


#endif /* JS_SUPPORT */


/*
 *	ANSI C requires atleast one dummy function in an
 *	entirly conditional module.
 */
void JSDummyFunc()
{
	int x;

	x = 0;


	return;
}
