/*
                         Object Create Script Names

	Functions:

	int OCSIsGarbage(int ocsn_num)
	void OCSReclaimMemory()
	void OCSDeleteAll()
	int OCSReset(int ocsn_num)
	int OCSCreate(int type)
	int OCSCreateExplicit(int ocsn_num, int type)
	void OCSRecycle(int ocsn_num)
	int OCSGetTop()
	int OCSGetByCode(int code)

	---

 */

#include "xsw.h"


/*
 *	Checks if OCSN is unallocated or garbage.
 */
int OCSIsGarbage(int ocsn_num)
{
        if((ocsn_num < 0) ||
           (ocsn_num >= total_ocsns) ||
           (ocsn == NULL)
        )
            return(-1);
        else if(ocsn[ocsn_num] == NULL)
            return(-1);
        else if(ocsn[ocsn_num]->type <= OCS_TYPE_GARBAGE)
            return(1);
        else
            return(0);
}


/*
 *	Free unused entries and reallocate pointers.
 */
void OCSReclaimMemory()
{
	int i, h;
	ocsn_struct **ocsn_ptr;


	/* Sanitize total. */
	if(total_ocsns < 0)
	    total_ocsns = 0;


        /* Get highest allocated ocsn. */
	for(i = 0, ocsn_ptr = ocsn, h = -1;
            i < total_ocsns;
            i++, ocsn_ptr++
        )
	{
	    if(*ocsn_ptr != NULL)
		h = i;
	}

	/* Adjust total and reallocate. */
	total_ocsns = h + 1;
	if(total_ocsns > 0)
	{
	    ocsn = (ocsn_struct **)realloc(
		ocsn,
		total_ocsns * sizeof(ocsn_struct *)
	    );
	    if(ocsn == NULL)
	    {
		total_ocsns = 0;
		return;
	    }
	}
	else
	{
	    total_ocsns = 0;

	    free(ocsn);
	    ocsn = NULL;
	}


	return;
}



/*
 *	Procedure to delete all OCSNs.
 */
void OCSDeleteAll()
{
	int i;


#ifdef DEBUG_MEM_FREE
printf("OCSN: Deleting %i...\n", total_ocsns);
#endif

        for(i = 0; i < total_ocsns; i++)
        {
            OCSReset(i);

#ifdef DEBUG_MEM_FREE
if(ocsn[i] != NULL)
    printf("OCSN %i: Free'ed.\n", i);
#endif
            free(ocsn[i]);
        }

#ifdef DEBUG_MEM_FREE
if(ocsn != NULL)
    printf("OCSN pointers: Free'ed.\n");
#endif
        free(ocsn);
        ocsn = NULL;

        total_ocsns = 0;


        return;
}



/*
 *	Frees all allocated substructures of ocsn.
 */
int OCSReset(int ocsn_num)
{
	if(OCSIsGarbage(ocsn_num))
	    return(-1);

        /* Reset values. */
	ocsn[ocsn_num]->type = OCS_TYPE_GARBAGE;
        ocsn[ocsn_num]->name[0] = '\0';


        return(0);
}



/*
 *	Allocates a new OCSN, returns the OCSN's number.
 */
int OCSCreate(int type)
{
        int ocsn_num;
        int ocs_rtn;


        /* ******************************************************** */

        /* Type may not be garbage. */
        if(type <= OCS_TYPE_GARBAGE)
        {
            fprintf(stderr,
                "OCSCreate: Error: Useless request to create garbage.\n"
            );
            return(-1);
        }


        /* ********************************************************* */

        /* Look for available OCS already allocated. */
        ocs_rtn = -1;
        for(ocsn_num = 0; ocsn_num < total_ocsns; ocsn_num++)
        {
            /* Look for garbage OCSs. */
            if(ocsn[ocsn_num]->type == OCS_TYPE_GARBAGE)
            {
                ocs_rtn = ocsn_num;
                break;
            }
        }

        /* Did we find an OCS already allocated? */
        if((ocs_rtn > -1) && (ocs_rtn < total_ocsns))
        {
            OCSReset(ocs_rtn);

            ocsn[ocs_rtn]->type = type;
	    ocsn[ocs_rtn]->name[0] = '\0';

            return(ocs_rtn);
        }


        /* Not enough memory, can we allocate more? */
        if((total_ocsns + OBJECT_ALLOCATE_AHEAD) > OCSN_MAX)
        {
            fprintf(stderr, "OCS maximum %i reached.\n", OCSN_MAX);
            return(-1);
        }


        /* Allocate more memory. */

        /* Allocate pointer array. */
        ocsn = (ocsn_struct **)realloc(
	    ocsn, (total_ocsns + OBJECT_ALLOCATE_AHEAD) *
            sizeof(ocsn_struct *)
	);
        if(ocsn == NULL)
        {
	    total_ocsns = 0;

            return(-1);
        }   


        /* Allocate each ocs. */
        for(ocsn_num = total_ocsns;
            ocsn_num < (total_ocsns + OBJECT_ALLOCATE_AHEAD);
            ocsn_num++
        )
        {
            ocsn[ocsn_num] = (ocsn_struct *)
                calloc(1, sizeof(ocsn_struct));

            /* Allocation error? */
            if(ocsn[ocsn_num] == NULL)
            {
                fprintf(stderr,
                    "OCS #%i: Memory allocation error.\n",
                    ocsn_num
                );
                return(-1);
            }
        }


        /* New object will be the first OCS allocated. */
        ocs_rtn = total_ocsns;


        /* Adjust global variable total_ocsns. */
	ocsn_num = total_ocsns;
        total_ocsns += OBJECT_ALLOCATE_AHEAD;

	/* Reset values on newly allocated objects. */
	while(ocsn_num < total_ocsns)
	{
	    OCSReset(ocsn_num);
	    ocsn_num++;
	}


        /* Set default values on the new OCS. */
        ocsn[ocs_rtn]->type = type;
	ocsn[ocs_rtn]->name[0] = '\0';


        /* Return the number of the newly created OCS. */
        return(ocs_rtn);
}



/*
 *	Allocates an OCSN explicitly, returns non-zero on error.
 */
int OCSCreateExplicit(int ocsn_num, int type)
{
        int ocs_count;
        int ocs_rtn;


        /* ********************************************************* */
        /* Type may not be garbage. */
        if(type <= OCS_TYPE_GARBAGE)
        {
            fprintf(stderr,
         "OCSCreateExplicit(): Error: Useless request to create garbage.\n"
            );
            return(-1);
        }

	/* Make sure ocsn_num is valid. */
	else if(ocsn_num < 0)
        {
            fprintf(stderr,
           "OCSCreateExplicit(): Error: Request to create ocsn_num #%i.\n",
		ocsn_num
            );
            return(-1);
	}
        
         
        /* *********************************************************** */

        /* If ocsn_num already exists, recycle it. */
        if(!OCSIsGarbage(ocsn_num))
            OCSRecycle(ocsn_num);


        /* Allocate memory if ocsn_num is greater than total_ocsns. */
        if(ocsn_num >= total_ocsns)
        {
            /* Can we allocate more? */
            if((ocsn_num + OBJECT_ALLOCATE_AHEAD) >= OCSN_MAX)
            {
                fprintf(stderr,
                    "OCSCreateExplicit: Maximum of %i OCSNs reached.\n",
		    OCSN_MAX
		);
                return(-1);
            }


            /* Allocate pointer array. */
            ocsn = (ocsn_struct **)realloc(
		ocsn,
		(ocsn_num + OBJECT_ALLOCATE_AHEAD + 1) * sizeof(ocsn_struct *)
            );
            if(ocsn == NULL)
            {
		total_ocsns = 0;
                return(-1);
            }


            /* Allocate each OCSN. */
            for(ocs_count = total_ocsns;
                ocs_count <= (ocsn_num + OBJECT_ALLOCATE_AHEAD);
                ocs_count++
            )
            {
                ocsn[ocs_count] = (ocsn_struct *)
                    calloc(1, sizeof(ocsn_struct));
         
                /* Allocation error? */
                if(ocsn[ocs_count] == NULL)
                {
                    fprintf(stderr,
                        "OCS #%i: Memory allocation error.\n",
                        ocs_count
                    );
                    return(-1);
                }
            }

            /* Adjust global variable total_ocsns. */
            ocs_count = total_ocsns;
            total_ocsns = ocsn_num + OBJECT_ALLOCATE_AHEAD + 1;

            while(ocs_count < total_ocsns)
            {
                OCSReset(ocs_count);
                ocs_count++;
            }
        }
             
                
        /* Create explicit OCS. */
        ocs_rtn = ocsn_num;

        /* Set default values on the new OCS. */
        OCSReset(ocs_rtn);

        ocsn[ocs_rtn]->type = type;
        ocsn[ocs_rtn]->name[0] = '\0';


        return(0);
}



/*
 *	Recycles ocsn, freeing all substructures and resources.
 *
 *	Does not free ocsn entry itself.
 */
void OCSRecycle(int ocsn_num)
{
	/* Reset values and free substructures. */
	OCSReset(ocsn_num);


	return;
}



int OCSGetTop()
{
        int i, last_valid_ocs;


	/* Get highest allocated and nongarbage OCSN. */
	last_valid_ocs = -1;
	for(i = 0; i < total_ocsns; i++)
	{
	    if(ocsn[i] == NULL)
	    {
		fprintf(stderr,
		    "OCSGetTop(): Error: OCSN #%i not allocated.\n",
		    i
		);
		continue;
	    }

	    if(ocsn[i]->type <= OCS_TYPE_GARBAGE)
		continue;

	    last_valid_ocs = i;
	}


	return(last_valid_ocs);
}



/*
 *	Returns an OCSN number by code (aka type).
 *	The code cannot be OCS_TYPE_GARBAGE.
 */
int OCSGetByCode(int code)
{
	int i;
	ocsn_struct **ocs_ptr;


	/* OCS code cannot be garbage. */
	if(code <= OCS_TYPE_GARBAGE)
            return(-1);


	/* Begin search.. */
	for(i = 0, ocs_ptr = ocsn;
	    i < total_ocsns;
	    i++, ocs_ptr++
	)
	{
	    if(*ocs_ptr == NULL)
		continue;

	    if((*ocs_ptr)->type == code)
		return(i);
	}


	return(-1);
}
