/*
                            Universe file IO

	Functions:

	char *UNVGetObjectFormalName(
		xsw_object_struct *obj_ptr,
		int obj_num
	)
	xsw_object_struct *UNVDupObject(xsw_object_struct *obj_ptr)
	void UNVResetObject(xsw_object_struct *obj_ptr)
	void UNVDeleteObject(xsw_object_struct *obj_ptr)
	void UNVDeleteAllObjects(xsw_object_struct **obj_ptr, int total)

	xsw_object_struct UNVLoadFromFile(
	        char *file,
	        int *total,
	        unv_head_struct *header_buf,
		void (*progress_notify)(int, int)
	)
	int UNVSaveToFile(
	        char *file,
	        xsw_object_struct **obj_ptr,
	        int total,
	        unv_head_struct *header_buf,
		void (*progress_notify)(int, int)
	)

	---


 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <db.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>

#include <math.h>

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/cfgfmt.h"
#include "../include/fio.h"
#include "../include/reality.h"
#include "../include/objects.h"
#include "../include/dbmatch.h"
#include "../include/unvfile.h"




int UNVAllocScores(xsw_object_struct *obj_ptr);
int UNVAllocObjectWeapons(xsw_object_struct *obj_ptr, int total_weapons);
int UNVAllocEco(xsw_object_struct *obj_ptr);




/*
 *	Allocates a score structure as needed on object.
 */
int UNVAllocScores(xsw_object_struct *obj_ptr)
{
	if(obj_ptr == NULL)
	    return(-1);

	if(obj_ptr->score == NULL)
	{
	    obj_ptr->score = (xsw_score_struct *)calloc(
		1,
                sizeof(xsw_score_struct)
	    );
	    if(obj_ptr->score == NULL)
            {
                return(-1);
            }
	}

	return(0);
}

/*
 *	Allocates or deallocates weapons on object as needed.
 */
int UNVAllocObjectWeapons(xsw_object_struct *obj_ptr, int total_weapons)
{
        int i, prev;


        if(obj_ptr == NULL)
	    return(-1);

        /* Deallocate all weapons? */
        if(total_weapons <= 0)
        {
            for(i = 0; i < obj_ptr->total_weapons; i++)
                free(obj_ptr->weapons[i]);

            free(obj_ptr->weapons);
            obj_ptr->weapons = NULL;

            obj_ptr->total_weapons = 0;
            obj_ptr->selected_weapon = -1;

            return(0);
        }


        /* Sanitize requested total_weapons. */
        if(total_weapons > MAX_WEAPONS)
            total_weapons = MAX_WEAPONS;

        /* Sanitize total_weapons on object. */
        if(obj_ptr->total_weapons > MAX_WEAPONS)
            obj_ptr->total_weapons = MAX_WEAPONS;   
        if(obj_ptr->total_weapons < 0)
            obj_ptr->total_weapons = 0;

        /* Record old total weapons on object. */
	prev = obj_ptr->total_weapons;


        /* Allocate or deallocate total weapons on object. */
        if(prev == total_weapons)
        {
            /* Total weapons remains the same, do nothing. */
            return(0);
        }
        else if(prev > total_weapons)
        {   
            /* Deallocate weapons on object. */
            for(i = prev - 1;
                i >= total_weapons;
                i--
            )
            {
                /* Unselect weapon if it will be deleted. */
                if(obj_ptr->selected_weapon == i)
                    obj_ptr->selected_weapon = -1;

                free(obj_ptr->weapons[i]);
                obj_ptr->weapons[i] = NULL;
            }
            obj_ptr->weapons = (xsw_weapons_struct **)realloc(
                obj_ptr->weapons,
                total_weapons * sizeof(xsw_weapons_struct *)
            );
            if(obj_ptr->weapons == NULL)
            {
                obj_ptr->total_weapons = 0;
                obj_ptr->selected_weapon = -1;
                return(-1);
            }

            /* Adjust the object's total_weapons to reflect changes. */
            obj_ptr->total_weapons = total_weapons;
        }   
        else
        {
            /* Allocate more weapons on object. */
            obj_ptr->weapons = (xsw_weapons_struct **)realloc(
                obj_ptr->weapons,
                total_weapons * sizeof(xsw_weapons_struct *)
            );
            if(obj_ptr->weapons == NULL)
            {
                obj_ptr->total_weapons = 0;
                obj_ptr->selected_weapon = -1;
                return(-1);
            }

            for(i = prev; i < total_weapons; i++)
            {
                obj_ptr->weapons[i] = (xsw_weapons_struct *)calloc(
                    1,
                    sizeof(xsw_weapons_struct)
                );
                if(obj_ptr->weapons[i] == NULL)
                {
                    obj_ptr->total_weapons = 0;
                    obj_ptr->selected_weapon = -1;
                    return(-1);
                }
            }

            /* Adjust the object's total_weapons to reflect changes. */
            obj_ptr->total_weapons = total_weapons;
        }


        /* Sanitize selected weapon. */
        if(obj_ptr->selected_weapon >= obj_ptr->total_weapons)
            obj_ptr->selected_weapon = obj_ptr->total_weapons - 1;
        if(obj_ptr->selected_weapon < 0)
            obj_ptr->selected_weapon = -1;


        return(0);
}


/*
 *	Allocates an economy structure as needed.
 */
int UNVAllocEco(xsw_object_struct *obj_ptr)
{
        if(obj_ptr == NULL)
            return(-1);

        if(obj_ptr->eco == NULL)
        {
            obj_ptr->eco = (xsw_ecodata_struct *)calloc(
                1, 
                sizeof(xsw_ecodata_struct)
            );
            if(obj_ptr->eco == NULL)
            {
                return(-1);
            }
        }
        
        return(0);
}

/*
 *	Returns a statically allocated string containing the object's
 *	formal name.
 */
char *UNVGetObjectFormalName(xsw_object_struct *obj_ptr, int obj_num)
{
	char *strptr;
	static char name[XSW_OBJ_NAME_MAX + 80];


	if(obj_ptr == NULL)
	    return(XSW_OBJ_GARBAGE_NAME);

	if(obj_ptr->type <= XSW_OBJ_TYPE_GARBAGE)
	    return(XSW_OBJ_GARBAGE_NAME);

	if(obj_num < 0)
	{
	    /* Object index number is invalid, implies garbage. */
	    strncpy(name, XSW_OBJ_GARBAGE_NAME, XSW_OBJ_NAME_MAX);
	    name[XSW_OBJ_NAME_MAX - 1] = '\0';
	    return(name);
	}
	else if(obj_ptr->type <= XSW_OBJ_TYPE_GARBAGE)
	{
	    /* Object type is garbage. */
            strncpy(name, XSW_OBJ_GARBAGE_NAME, XSW_OBJ_NAME_MAX);
            name[XSW_OBJ_NAME_MAX - 1] = '\0';
            return(name);
	}

	sprintf(
	    name,
	    "%s(#%i",
	    obj_ptr->name,
	    obj_num
	);
	strptr = name;
	while(*strptr != '\0')
	    strptr++;

	/* Type. */
	switch(obj_ptr->type)
	{
	  case XSW_OBJ_TYPE_DYNAMIC:
	    *strptr++ = 'D';
	    break;

	  case XSW_OBJ_TYPE_CONTROLLED:
            *strptr++ = 'C';
            break;

          case XSW_OBJ_TYPE_PLAYER:
            *strptr++ = 'P';
            break;

          case XSW_OBJ_TYPE_WEAPON:
            *strptr++ = 'W';
            break;

          case XSW_OBJ_TYPE_STREAMWEAPON:
            *strptr++ = 'W';
            break;

          case XSW_OBJ_TYPE_SPHEREWEAPON:
            *strptr++ = 'W';
            break;

          case XSW_OBJ_TYPE_HOME:
            *strptr++ = 'H';
            break;

          case XSW_OBJ_TYPE_AREA:
            *strptr++ = 'A';
            break;

          case XSW_OBJ_TYPE_ANIMATED:
            *strptr++ = 'V';
            break;

          default:
            *strptr++ = '-';
            break;
	}

	if(obj_ptr->eco != NULL)
	    *strptr++ = 'E';

	*strptr++ = ')';

	/* Null terminate. */
        *strptr = '\0';


	return(name);
}


/*
 *	Dynamically allocates a new object in memory that contains
 *	the exact same values (except for pointers to substructures) as
 *	obj_ptr.  Can return NULL on error.
 */
xsw_object_struct *UNVDupObject(xsw_object_struct *obj_ptr)
{
	int i;
	xsw_object_struct *new_obj_ptr = NULL;


	if(obj_ptr == NULL)
	    return(new_obj_ptr);


	/* Allocate new. */
	new_obj_ptr = (xsw_object_struct *)calloc(
	    1,
	    sizeof(xsw_object_struct)
	);
	if(new_obj_ptr == NULL)
	    return(new_obj_ptr);


	/* Copy over core values. */
        memcpy(
            new_obj_ptr,
            obj_ptr,
            sizeof(xsw_object_struct)
        );


	/* Copy over substructures. */

	/* Weapons. */
	if(obj_ptr->weapons != NULL)
	{
	    /* Copy over each weapon. */
	    new_obj_ptr->weapons = (xsw_weapons_struct **)calloc(
		1,
		new_obj_ptr->total_weapons * sizeof(xsw_weapons_struct *)
	    );
	    if(new_obj_ptr->weapons == NULL)
	    {
		new_obj_ptr->total_weapons = 0;
	    }

	    for(i = 0; i < new_obj_ptr->total_weapons; i++)
	    {
		if(obj_ptr->weapons[i] == NULL)
		    continue;

		new_obj_ptr->weapons[i] = (xsw_weapons_struct *)calloc(
		    1,
		    sizeof(xsw_weapons_struct)
		);
		if(new_obj_ptr->weapons[i] == NULL)
		    continue;

		memcpy(
		    new_obj_ptr->weapons[i],
		    obj_ptr->weapons[i],
		    sizeof(xsw_weapons_struct)
		);
	    }
	}

	/* Score. */
        if(obj_ptr->score != NULL)
        {
            new_obj_ptr->score = (xsw_score_struct *)malloc(
                sizeof(xsw_score_struct)
            );
            if(new_obj_ptr->score != NULL)
            {
		memcpy(
		    new_obj_ptr->score,
		    obj_ptr->score,
		    sizeof(xsw_score_struct)
		);
	    }
	}

        /* Economy. */
        if(obj_ptr->eco != NULL)
        {
            new_obj_ptr->eco = (xsw_ecodata_struct *)malloc(
                sizeof(xsw_ecodata_struct)
            );
            if(new_obj_ptr->eco != NULL)  
            {
                memcpy(
                    new_obj_ptr->eco,
                    obj_ptr->eco,
                    sizeof(xsw_ecodata_struct)
                );
            }

	    /* Copy over each eco product. */
	    if(obj_ptr->eco->product != NULL)
	    {
		new_obj_ptr->eco->product = (xsw_ecoproduct_struct **)calloc(
		    1,
		    new_obj_ptr->eco->total_products * sizeof(xsw_ecoproduct_struct *)
		);
		if(new_obj_ptr->eco->product == NULL)
		{
		    new_obj_ptr->eco->total_products = 0;
		}

                for(i = 0; i < new_obj_ptr->eco->total_products; i++)
                {
                    if(obj_ptr->eco->product[i] == NULL)
                        continue;

                    new_obj_ptr->eco->product[i] = (xsw_ecoproduct_struct *)calloc(
                        1,
                        sizeof(xsw_ecoproduct_struct)
                    );
                    if(new_obj_ptr->eco->product[i] == NULL)
                        continue;

                    memcpy(
                        new_obj_ptr->eco->product[i],
                        obj_ptr->eco->product[i],
                        sizeof(xsw_ecoproduct_struct)
                    );
		}
	    }
        }



	return(new_obj_ptr);
}


/*
 *	Resets values and deallocates all substructures on
 *	object.
 */
void UNVResetObject(xsw_object_struct *obj_ptr)
{
        int i;


        if(obj_ptr == NULL)
            return;


        /* Deallocate weapons. */
        for(i = 0; i < obj_ptr->total_weapons; i++)
        {
            if(obj_ptr->weapons[i] == NULL)
                continue;

            free(obj_ptr->weapons[i]);
        }
        free(obj_ptr->weapons);
        obj_ptr->weapons = NULL;

        obj_ptr->total_weapons = 0;
        obj_ptr->selected_weapon = -1;

 
        /* Deallocate tractored objects. */
        free(obj_ptr->tractored_object);
        obj_ptr->tractored_object = NULL;
        obj_ptr->total_tractored_objects = 0;


        /* Deallocate score structure. */
        free(obj_ptr->score);
        obj_ptr->score = NULL;


        /* Deallocate economy structure. */
        if(obj_ptr->eco != NULL)
        {           
            for(i = 0; i < obj_ptr->eco->total_products; i++)
            {
                free(obj_ptr->eco->product[i]);
                obj_ptr->eco->product[i] = NULL;
            }
            free(obj_ptr->eco->product);
            obj_ptr->eco->product = NULL;
        }
        free(obj_ptr->eco);
        obj_ptr->eco = NULL;


        /* Reset values to that of the garbage_object. */
        memcpy(
            obj_ptr,		/* Destination. */
            &garbage_object,	/* Source. */
            sizeof(xsw_object_struct)
        );


	return;
}


/*
 *	Deallocates object and all its substructures.
 */
void UNVDeleteObject(xsw_object_struct *obj_ptr)
{
        if(obj_ptr == NULL)
	    return;


        /* Free all allocated substructures. */
        UNVResetObject(obj_ptr);

        /* Free memory for givin object. */
        free(obj_ptr);
        obj_ptr = NULL;


	return;
}


/*
 *	Deletes an array of objects and their substructures,
 *	including the array itself.
 */
void UNVDeleteAllObjects(xsw_object_struct **obj_ptr, int total)
{
	int i;
	xsw_object_struct **ptr;


	/* Deallocate each object. */
	for(i = 0, ptr = obj_ptr;
            i < total;
            i++, ptr++
	)
	    UNVDeleteObject(*ptr);

	/* Deallocate object pointer array. */
	free(obj_ptr);


	return;
}



/*
 *	Loads objects from file.  The loaded objects are dynamically
 *	allocated and need to be deallocated with a call to
 *	UNVDeleteAllObjects().
 */
xsw_object_struct **UNVLoadFromFile(
        char *file,
        int *total,
	unv_head_struct *header_buf,
	void *client_data,
	void (*progress_notify)(void *, int, int)
)
{
	int i, n, prev, status;
	char *strptr, *strptr2;
        off_t filesize;
	FILE *fp;
	struct stat stat_buf;

	char parm[CFG_PARAMETER_MAX];
	char val[CFG_VALUE_MAX];

	int lines_read = 0;

	int object_num = -1;
	int weapons_count = 0;
	int product_num = 0;

	xsw_ecodata_struct *eco_ptr;
	xsw_object_struct *obj_ptr;
	xsw_object_struct **ptr = NULL;



	if(total == NULL)
	    return(ptr);

	*total = 0;

	if(file == NULL)
	    return(ptr);
        if(stat(file, &stat_buf))
        {
            fprintf(stderr,
                "%s: No such file.\n",
                file
            );
            return(ptr);
        }

	/* Get size of file. */
        filesize = stat_buf.st_size;

	/* Open file. */
	fp = fopen(file, "r");
	if(fp == NULL)
	{
            fprintf(stderr,
		"%s: Cannot open.\n",
                file
            );
            return(ptr);
	}


	/* Reset header. */
	if(header_buf != NULL)
	{
            strncpy(header_buf->title, "Untitled", UNV_TITLE_MAX);
            header_buf->title[UNV_TITLE_MAX - 1] = '\0';

            header_buf->isr[0] = '\0';
            header_buf->ocsn[0] = '\0';
            header_buf->ss[0] = '\0';
	}


	/* ************************************************ */

	/* Call notify progress function. */
	if(progress_notify != NULL)
	    progress_notify(
		client_data,
		(int)ftell(fp),
		(int)filesize
	    );


        strptr = NULL;

	while(1)
	{
            /* Free previous line and allocate/read next line. */
            free(strptr); strptr = NULL;
            strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
            if(strptr == NULL) break;
            lines_read++;

            /* Fetch parameter. */
            strptr2 = StringCfgParseParm(strptr);
            if(strptr2 == NULL) continue;
            strncpy(parm, strptr2, CFG_PARAMETER_MAX);
            parm[CFG_PARAMETER_MAX - 1] = '\0';

            /* Fetch value. */
            strptr2 = StringCfgParseValue(strptr);
	    /* Set value to "0" if NULL pointer was returned. */
            if(strptr2 == NULL) strptr2 = "0";
            strncpy(val, strptr2, CFG_VALUE_MAX);
            val[CFG_VALUE_MAX - 1] = '\0';


            /* Call notify progress function. */
            if(progress_notify != NULL)
                progress_notify(
                    client_data,
                    (int)ftell(fp),
                    (int)filesize
                );


	    /* ***************************************************** */
	    /* BeginHeader */
	    if(!strcmp(parm, "BeginHeader"))
	    {
                while(1)
                {
                    /* Free previous line and allocate/read next line. */
                    free(strptr); strptr = NULL;
                    strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
                    if(strptr == NULL) break;
                    lines_read++;

                    /* Fetch parameter. */
                    strptr2 = StringCfgParseParm(strptr);
                    if(strptr2 == NULL) continue;
                    strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                    parm[CFG_PARAMETER_MAX - 1] = '\0';

                    /* Fetch value. */
                    strptr2 = StringCfgParseValue(strptr);
                    /* Set value to "0" if NULL pointer was returned. */ 
                    if(strptr2 == NULL) strptr2 = "0";
                    strncpy(val, strptr2, CFG_VALUE_MAX);
                    val[CFG_VALUE_MAX - 1] = '\0';


                    /* Title */
                    if(!strcmp(parm, "Title"))
                    {
			if(header_buf != NULL)
			{
                            strncpy(
				header_buf->title,
				val,
				UNV_TITLE_MAX
			    );
			    header_buf->title[UNV_TITLE_MAX - 1] = '\0';
			}
                    }
		    /* Version. */
                    else if(!strcmp(parm, "Version"))
                    {
                        if(header_buf != NULL)
                        {
                            strncpy(
                                header_buf->version,
                                val,
                                UNV_TITLE_MAX
                            );
                            header_buf->version[UNV_TITLE_MAX - 1] = '\0';
                        }
                    }
		    /* UnitConversionRUTOAU */
		    else if(!strcmp(parm, "UnitConversionRUTOAU"))
		    {
                        if(header_buf != NULL)
                        {
			    header_buf->ru_to_au = atof(val);
			    if(header_buf->ru_to_au <= 0)
			    {
			        fprintf(stderr,
 "%s: Line %i: UnitConversionRUTOAU `%s' out of range.\n",
                                    file, lines_read, val
                                );
			        header_buf->ru_to_au = DEF_UNITCONV_RU_TO_AU;
			    }
		        }
		    }
                    /* ImageSetFile */
                    else if(!strcmp(parm, "ImageSetFile"))
                    {
                        if(header_buf != NULL)
                        {
                            strncpy(header_buf->isr, val, PATH_MAX + NAME_MAX);
                            header_buf->isr[PATH_MAX + NAME_MAX - 1] = '\0';
			}
                    }
                    /* OCSNFile */
                    else if(!strcmp(parm, "OCSNFile"))
                    {
                        if(header_buf != NULL)
                        {
                            strncpy(header_buf->ocsn, val, PATH_MAX + NAME_MAX);
                            header_buf->ocsn[PATH_MAX + NAME_MAX - 1] = '\0';
			}
                    }
                    /* SoundSchemeFile */
                    else if(!strcmp(parm, "SoundSchemeFile"))
                    {
			if(header_buf != NULL)
                        {
                            strncpy(header_buf->ss, val, PATH_MAX + NAME_MAX);
                            header_buf->ss[PATH_MAX + NAME_MAX - 1] = '\0';
                        }
		    }
		    /* LostFoundOwner */
		    else if(!strcmp(parm, "LostFoundOwner"))
		    {
                        if(header_buf != NULL)
                        {
                            header_buf->lost_found_owner = atol(val);
			}
		    }

                    /* EndHeader */
                    else if(!strcmp(parm, "EndHeader"))
                    {
                        break;
                    }
                }
		continue;
	    }
            /* ***************************************************** */
            /* BeginXSWObject */
	    else if(!strcmp(parm, "BeginXSWObject") ||
	            !strcmp(parm, "BeginObject")
	    )
	    {
		/* Get object number. */
		object_num = atol(val);
		if((object_num < 0) || (object_num >= MAX_OBJECTS))
		{
                    fprintf(stderr,
           "%s: Line %i: Object number `%i' out of range.\n",
			file, lines_read, object_num
		    );
		    continue;
                }

		/* Allocate a new object in memory. */

		/* Need to allocate more object pointers? */
		if(object_num >= *total)
		{
		    prev = *total;
		    *total = object_num + 1;

		    ptr = (xsw_object_struct **)realloc(
			ptr,
			*total * sizeof(xsw_object_struct *)
		    );
		    if(ptr == NULL)
		    {
			*total = 0;
			continue;
		    }

		    /* Allocate each new object. */
		    for(i = prev; i < *total; i++)
		    {
			ptr[i] = (xsw_object_struct *)calloc(
                            1,
                            sizeof(xsw_object_struct)
                        );
			UNVResetObject(ptr[i]);
		    }
		}
		/* Allocate object structure as needed. */
		else if(ptr[object_num] == NULL)
		{
		    ptr[object_num] = (xsw_object_struct *)calloc(
			1,
			sizeof(xsw_object_struct)
		    );
		    if(ptr[object_num] == NULL)
		    {
			continue;
		    }
		}
		else
		{
                    fprintf(stderr,
              "%s: Line %i: Warning: Redefining object #%i.\n",
                        file, lines_read, object_num
                    );
		}

		/* Get pointer to object. */
		obj_ptr = ptr[object_num];

		UNVResetObject(obj_ptr);

		/* Reset weapons count. */
		weapons_count = 0;

		while(1)
                {
                    /* Free previous line and allocate/read next line. */
                    free(strptr); strptr = NULL;
                    strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
                    if(strptr == NULL) break;
                    lines_read++;

                    /* Fetch parameter. */
                    strptr2 = StringCfgParseParm(strptr);
                    if(strptr2 == NULL) continue;
                    strncpy(parm, strptr2, CFG_PARAMETER_MAX);   
                    parm[CFG_PARAMETER_MAX - 1] = '\0';

                    /* Fetch value. */
                    strptr2 = StringCfgParseValue(strptr);
                    /* Set value to "0" if NULL pointer was returned. */
                    if(strptr2 == NULL) strptr2 = "0";
                    strncpy(val, strptr2, CFG_VALUE_MAX);
                    val[CFG_VALUE_MAX - 1] = '\0';


		    /* Name. */
		    if(!strcmp(parm, "Name"))
		    {
                        strncpy(
			    obj_ptr->name,
			    val,
			    XSW_OBJ_NAME_MAX
			);
			obj_ptr->name[XSW_OBJ_NAME_MAX - 1] = '\0';
		    }
                    /* Password. */
                    else if(!strcmp(parm, "Password"))
                    {
                        strncpy(
			    obj_ptr->password,
			    val,
			    XSW_OBJ_PASSWORD_MAX
			);
			obj_ptr->password[
			    XSW_OBJ_PASSWORD_MAX - 1] = '\0';
                    }
                    /* Empire. */
                    else if(!strcmp(parm, "Empire"))
                    {
			strncpy(
			    obj_ptr->empire,
			    val,
			    XSW_OBJ_EMPIRE_MAX
			);
			obj_ptr->empire[
			    XSW_OBJ_EMPIRE_MAX - 1] = '\0';
                    }
                    /* Type */  
                    else if(!strcmp(parm, "Type"))
                    {
			obj_ptr->type = atoi(val);

			/* Type must be a non-garbage. */
			if(obj_ptr->type <= XSW_OBJ_TYPE_GARBAGE)
			{
			    obj_ptr->type =
			        XSW_OBJ_TYPE_STATIC;
			}
		    }
                    /* LocType */
                    else if(!strcmp(parm, "LocType"))
                    {
                        obj_ptr->loc_type = atoi(val);

                        /* Location type must be valid. */
                        if(obj_ptr->loc_type <
			    XSW_LOC_TYPE_SPACE)
                        {
                            obj_ptr->loc_type =
                                XSW_LOC_TYPE_SPACE;
                        }
                    }
                    /* ImageSet */
                    else if(!strcmp(parm, "ImageSet"))
                    {
                        obj_ptr->imageset = atoi(val);
                    }
                    /* Owner */
                    else if(!strcmp(parm, "Owner"))
                    {
                        obj_ptr->owner = atol(val);
			if(obj_ptr->owner >= MAX_OBJECTS)
			    obj_ptr->owner = -1;
                    }
                    /* Size */
                    else if(!strcmp(parm, "Size"))
                    {
                        obj_ptr->size = atol(val);
			if(obj_ptr->size < 1)
			    obj_ptr->size = 1;
                    }
                    /* LockedOn */
                    else if(!strcmp(parm, "LockedOn"))
                    {
			obj_ptr->locked_on = atol(val);
                        if(obj_ptr->locked_on >= MAX_OBJECTS)
                            obj_ptr->locked_on = -1;
                    }
                    /* InterceptingObject */
                    else if(!strcmp(parm, "InterceptingObject"))
                    {
                        obj_ptr->intercepting_object =
			    atol(val);
                        if(obj_ptr->intercepting_object >=
                            MAX_OBJECTS)
                            obj_ptr->intercepting_object = -1;
                    }
		    /* ScannerRange */
                    else if(!strcmp(parm, "ScannerRange"))
                    {
                        obj_ptr->scanner_range =
			    atof(val);
			if(obj_ptr->scanner_range < 1)
			    obj_ptr->scanner_range = 1;
                    }
		    /* SectX */
                    else if(!strcmp(parm, "SectX"))
                    {
                        obj_ptr->sect_x = atol(val);
                    }
                    /* SectY */
                    else if(!strcmp(parm, "SectY"))
                    {
                        obj_ptr->sect_y = atol(val);
                    }
                    /* SectZ */
                    else if(!strcmp(parm, "SectZ"))
                    {
                        obj_ptr->sect_z = atol(val);
                    }
                    /* CoordX */
                    else if(!strcmp(parm, "CoordX"))
                    {
                        obj_ptr->x = atof(val);
                    }
                    /* CoordY */
                    else if(!strcmp(parm, "CoordY"))
                    {
                        obj_ptr->y = atof(val);
                    }
                    /* CoordZ */
                    else if(!strcmp(parm, "CoordZ"))
                    {
                        obj_ptr->z = atof(val);
                    }
                    /* ObjectHeading (in radians). */
                    else if(!strcmp(parm, "ObjectHeading"))
                    {
                        obj_ptr->heading = atof(val);
                    }
                    /* ObjectPitch (in radians). */
                    else if(!strcmp(parm, "ObjectPitch"))
                    {
                        obj_ptr->pitch = atof(val);
                    }
                    /* ObjectBank (in radians). */
                    else if(!strcmp(parm, "ObjectBank"))
                    {
                        obj_ptr->bank = atof(val);
                    }
                    /* VelocityHeading (in radians) */
                    else if(!strcmp(parm, "VelocityHeading"))
                    {
                        obj_ptr->velocity_heading = atof(val);
                    }
                    /* VelocityMax */
                    else if(!strcmp(parm, "VelocityMax"))
                    {
                        obj_ptr->velocity_max = atof(val);
                    }
                    /* Velocity */
                    else if(!strcmp(parm, "Velocity"))
                    {
                        obj_ptr->velocity = atof(val);
                    }
                    /* ThrustDir */
                    else if(!strcmp(parm, "ThrustDir"))
                    {
                        obj_ptr->thrust_dir = atof(val);
                    }
                    /* ThrustPower */
                    else if(!strcmp(parm, "ThrustPower"))
                    {
                        obj_ptr->thrust_power = atof(val);
                    }
		    /* Thrust */
                    else if(!strcmp(parm, "Thrust"))
                    {
                        obj_ptr->thrust = atof(val);
                    }
                    /* Throttle */
                    else if(!strcmp(parm, "Throttle"))
                    {
                        obj_ptr->throttle = atof(val);
                    }
                    /* EngineState */
                    else if(!strcmp(parm, "EngineState"))
                    {
                        obj_ptr->engine_state = atoi(val);
                    }
                    /* TurnRate */
                    else if(!strcmp(parm, "TurnRate"))
                    {
                        obj_ptr->turnrate = atof(val);
                    }
                    /* Lighting */
                    else if(!strcmp(parm, "Lighting"))
                    {
                        obj_ptr->lighting = atoi(val);
                    }
                    /* BrakesState */
                    else if(!strcmp(parm, "BrakesState"))
                    {
			/* No longer supported. */
                    }
                    /* BrakesPower */
                    else if(!strcmp(parm, "BrakesPower"))
                    {
                        /* No longer supported. */
                    }
		    /* HitPointsMax */
                    else if(!strcmp(parm, "HitPointsMax"))
                    {
                        obj_ptr->hp_max = atof(val);
                    }
                    /* HitPoints */
                    else if(!strcmp(parm, "HitPoints"))
                    {
                        obj_ptr->hp = atof(val);
                    }
                    /* PowerPurity */
                    else if(!strcmp(parm, "PowerPurity"))
                    {
                        obj_ptr->power_purity = atof(val);
                    }
                    /* PowerMax */
                    else if(!strcmp(parm, "PowerMax"))
                    {
                        obj_ptr->power_max = atof(val);
                    }
                    /* Power */
                    else if(!strcmp(parm, "Power"))
                    {
                        obj_ptr->power = atof(val);
                    }
                    /* CoreEfficency */
                    else if(!strcmp(parm, "CoreEfficency"))
                    {
                        obj_ptr->core_efficency = atof(val);
                    }
                    /* AntimatterMax */
                    else if(!strcmp(parm, "AntimatterMax"))
                    {
                        obj_ptr->antimatter_max = atof(val);
                    }
                    /* Antimatter */
                    else if(!strcmp(parm, "Antimatter"))
                    {
                        obj_ptr->antimatter = atof(val);
                    }
                    /* ShieldState */
                    else if(!strcmp(parm, "ShieldState"))
                    {
                        obj_ptr->shield_state = atoi(val);
                    }
                    /* ShieldFrequency */
                    else if(!strcmp(parm, "ShieldFrequency"))
                    {
                        obj_ptr->shield_frequency = atof(val);
                    }
		    /* SelectedWeapon */
                    else if(!strcmp(parm, "SelectedWeapon"))
                    {
                        obj_ptr->selected_weapon = atoi(val);
                    }
                    /* TotalWeapons */
                    else if(!strcmp(parm, "TotalWeapons"))
                    {
			if(UNVAllocObjectWeapons(obj_ptr, atoi(val)))
			{
                            fprintf(stderr,
                        "%s: Line %i: Error allocating %s weapons.\n",
                                file, lines_read, val
                            );

			    obj_ptr->total_weapons = 0;
			    obj_ptr->selected_weapon = -1;
			}
                    }
                    /* BirthTime */
                    else if(!strcmp(parm, "BirthTime"))
                    {
                        obj_ptr->birth_time = atol(val);
                    }
                    /* LifeSpan */
                    else if(!strcmp(parm, "LifeSpan"))
                    {
                        obj_ptr->lifespan = atol(val);
                    }
                    /* CloakState */
                    else if(!strcmp(parm, "CloakState"))
                    {
                        obj_ptr->cloak_state = atoi(val);
                    }
                    /* CloakStrength */
                    else if(!strcmp(parm, "CloakStrength"))
                    {
                        obj_ptr->cloak_strength = atof(val);
			if(obj_ptr->cloak_strength < 0)
			    obj_ptr->cloak_strength = 0;
                    }
                    /* ShieldVisibility */
                    else if(!strcmp(parm, "ShieldVisibility"))
                    {
                        obj_ptr->shield_visibility = atof(val);
			if(obj_ptr->shield_visibility < 0)
			    obj_ptr->shield_visibility = 0;
                    }
                    /* CurrentVisibility */
                    else if(!strcmp(parm, "CurrentVisibility"))
                    {
                        obj_ptr->cur_visibility = atof(val);
			if(obj_ptr->cur_visibility < 0)
			    obj_ptr->cur_visibility = 0;
                    }
                    /* Visibility */
                    else if(!strcmp(parm, "Visibility"))
                    {
                        obj_ptr->visibility = atof(val);
			if(obj_ptr->visibility < 0)
			    obj_ptr->visibility = 0;
                    }
                    /* DamageControl */
                    else if(!strcmp(parm, "DamageControl"))
                    {
                        obj_ptr->damage_control = atoi(val);
                    }
		    /* ComChannel */
                    else if(!strcmp(parm, "ComChannel"))
                    {
                        obj_ptr->com_channel = atoi(val);
			if(obj_ptr->com_channel < 0)
			    obj_ptr->com_channel = 0;
                    }

		    /* *************************************************** */
		    /* ScoreCredits */
		    else if(!strcmp(parm, "ScoreCredits"))
		    {
			if(!UNVAllocScores(obj_ptr))
			{
			    obj_ptr->score->credits = atof(val);
			}
		    }
                    /* ScoreRMUMax */
                    else if(!strcmp(parm, "ScoreRMUMax"))
                    {
                        if(!UNVAllocScores(obj_ptr))
                        {
                            obj_ptr->score->rmu_max = atof(val);
			}
                    }
                    /* ScoreRMU */
                    else if(!strcmp(parm, "ScoreRMU"))
                    {
                        if(!UNVAllocScores(obj_ptr))
                        {   
                            obj_ptr->score->rmu = atof(val);
			}
                    }
                    /* ScoreDamageGivin */
                    else if(!strcmp(parm, "ScoreDamageGivin"))
                    {
                        if(!UNVAllocScores(obj_ptr))
                        {
                            obj_ptr->score->damage_givin = atof(val);
			}
                    }
                    /* ScoreDamageRecieved */
                    else if(!strcmp(parm, "ScoreDamageRecieved"))
                    {
                        if(!UNVAllocScores(obj_ptr))
                        {   
                            obj_ptr->score->damage_recieved = atof(val);
			}
                    }
                    /* ScoreKills */
                    else if(!strcmp(parm, "ScoreKills"))
                    {
                        if(!UNVAllocScores(obj_ptr))
                        {
                            obj_ptr->score->kills = atol(val);
			}
                    }
                    /* *************************************************** */
                    /* PermissionUID */
                    else if(!strcmp(parm, "PermissionUID"))
                    {
                        obj_ptr->permission.uid = atoi(val);
                    }
                    /* PermissionGID */
                    else if(!strcmp(parm, "PermissionGID"))
                    {
                        obj_ptr->permission.gid = atoi(val);
                    }
                    /* *************************************************** */
                    /* AnimationLastInterval */
                    else if(!strcmp(parm, "AnimationLastInterval"))
                    {
                        obj_ptr->animation.last_interval = atol(val);
                    }
                    /* AnimationInterval */
                    else if(!strcmp(parm, "AnimationInterval"))
                    {
                        obj_ptr->animation.interval = atol(val);
                    }
                    /* AnimationCurrentFrame */
                    else if(!strcmp(parm, "AnimationCurrentFrame"))
                    {
                        obj_ptr->animation.current_frame = atol(val);
                    }
                    /* AnimationTotalFrames */
                    else if(!strcmp(parm, "AnimationTotalFrames"))
                    {
                        obj_ptr->animation.total_frames = atol(val);
                    }
                    /* AnimationCycleCount */
                    else if(!strcmp(parm, "AnimationCycleCount"))
                    {
                        obj_ptr->animation.cycle_count = atol(val);
                    }
                    /* AnimationCycleTimes */
                    else if(!strcmp(parm, "AnimationCycleTimes"))
                    {
                        obj_ptr->animation.cycle_times = atol(val);
                    }
		    /* **************************************************** */
		    /* BeginWeapon */
		    else if(!strcmp(parm, "BeginWeapon"))
		    {
			while(1)
                        {
			    /* Free previous line and allocate/read next line. */
                            free(strptr); strptr = NULL;
                            strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
                            if(strptr == NULL) break;
                            lines_read++;

                            /* Fetch parameter. */
                            strptr2 = StringCfgParseParm(strptr);
                            if(strptr2 == NULL) continue;
                            strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                            parm[CFG_PARAMETER_MAX - 1] = '\0';

                            /* Fetch value. */
                            strptr2 = StringCfgParseValue(strptr); 
                            /* Set value to "0" if NULL pointer was returned. */
                            if(strptr2 == NULL) strptr2 = "0";
                            strncpy(val, strptr2, CFG_VALUE_MAX);
                            val[CFG_VALUE_MAX - 1] = '\0';


                            /* EmissionType */
                            if(!strcmp(parm, "EmissionType"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->emission_type
			                = atoi(val);
                            }
                            /* OCSCode */
                            else if(!strcmp(parm, "OCSCode"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->ocs_code
                                        = atoi(val);
                            }
                            /* Amount */
                            else if(!strcmp(parm, "Amount"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->amount
                                        = atol(val);
                            }
                            /* Max */
                            else if(!strcmp(parm, "Max"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->max
                                        = atol(val);
                            }
                            /* CreatePower */
                            else if(!strcmp(parm, "CreatePower"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->create_power
                                        = atof(val);
                            }
                            /* Power */
                            else if(!strcmp(parm, "Power"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->power
                                        = atof(val);
                            }
                            /* Range */
                            else if(!strcmp(parm, "Range"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->range
                                        = atol(val);
                            }
                            /* Delay */
                            else if(!strcmp(parm, "Delay"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->delay
                                        = atol(val);
                            }
                            /* LastUsed */
                            else if(!strcmp(parm, "LastUsed"))
                            {
                                if(weapons_count < obj_ptr->total_weapons)
                                    obj_ptr->weapons[weapons_count]->last_used
                                        = atol(val);
			    }
			    /* EndWeapon */
			    else if(!strcmp(parm, "EndWeapon"))
			    {
			        weapons_count++;
			        break;
			    }
                            /* Unsupported property */
                            else
                            {
                                fprintf(stderr,
                      "%s: Line %i: Warning: Unsupported parameter `%s'\n",
                                    file, lines_read, parm
                                );
                            }
		        }
		    }
		    /* ************************************************ */
		    /* BeginEcoData */
                    else if(!strcmp(parm, "BeginEcoData"))
                    {
			/* Allocate economy structure on object as needed. */
		        if(UNVAllocEco(obj_ptr))
                            continue;

		        eco_ptr = obj_ptr->eco;
		        if(eco_ptr == NULL)
			    continue;

                        while(1)
                        {
                            /* Free previous line and allocate/read next line. */
                            free(strptr); strptr = NULL;
                            strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
                            if(strptr == NULL) break;
                            lines_read++;

                            /* Fetch parameter. */
                            strptr2 = StringCfgParseParm(strptr);
                            if(strptr2 == NULL) continue;
                            strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                            parm[CFG_PARAMETER_MAX - 1] = '\0';

                            /* Fetch value. */
                            strptr2 = StringCfgParseValue(strptr);
                            /* Set value to "0" if NULL pointer was returned. */
                            if(strptr2 == NULL) strptr2 = "0";          
                            strncpy(val, strptr2, CFG_VALUE_MAX);
                            val[CFG_VALUE_MAX - 1] = '\0';


                            /* Flags */
		            if(!strcmp(parm, "Flags"))
                            {
                                eco_ptr->flags = atol(val);
                            }
                            /* TaxGeneral */
                            else if(!strcmp(parm, "TaxGeneral"))
                            {
                                eco_ptr->tax_general = atof(val);
                                if(eco_ptr->tax_general < 0.00001) 
                                    eco_ptr->tax_general = 0.00001;
                            }
                            /* TaxFriend */
                            else if(!strcmp(parm, "TaxFriend"))
                            {
                                eco_ptr->tax_friend = atof(val);
                                if(eco_ptr->tax_friend < 0.00001)
                                    eco_ptr->tax_friend = 0.00001;
                            }
                            /* TaxHostile */
                            else if(!strcmp(parm, "TaxHostile"))
                            {
                                eco_ptr->tax_hostile = atof(val);
                                if(eco_ptr->tax_hostile < 0.00001)
                                    eco_ptr->tax_hostile = 0.00001;
                            }
			    /* **************************************** */
			    /* BeginEcoProduct */
                            else if(!strcmp(parm, "BeginEcoProduct"))
                            {
                                /* Allocate an eco product. */
                                if(eco_ptr->total_products < 0)
                                  eco_ptr->total_products = 0;

                                product_num = eco_ptr->total_products;
                                eco_ptr->total_products++;

                                eco_ptr->product = (xsw_ecoproduct_struct **)realloc(
                                  eco_ptr->product,
                                  eco_ptr->total_products * sizeof(xsw_ecoproduct_struct *)
                                );
                                if(eco_ptr->product == NULL)
                                {
                                  eco_ptr->total_products = 0;
                                  continue;
                                }

                                eco_ptr->product[product_num] = (xsw_ecoproduct_struct *)
                                    calloc(1, sizeof(xsw_ecoproduct_struct));
                                if(eco_ptr->product[product_num] == NULL)
                                {
                                  continue;
                                }

                                while(1)
                                {
                                  /* Free previous line and allocate/read next line. */
                                  free(strptr); strptr = NULL;
                                  strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
                                  if(strptr == NULL) break;
                                  lines_read++;

                                  /* Fetch parameter. */
                                  strptr2 = StringCfgParseParm(strptr);  
                                  if(strptr2 == NULL) continue;   
                                  strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                                  parm[CFG_PARAMETER_MAX - 1] = '\0';

                                  /* Fetch value. */
                                  strptr2 = StringCfgParseValue(strptr);
                                  /* Set value to "0" if NULL pointer was returned. */
                                  if(strptr2 == NULL) strptr2 = "0";
                                  strncpy(val, strptr2, CFG_VALUE_MAX);
                                  val[CFG_VALUE_MAX - 1] = '\0';


                                  /* Name */
                                  if(!strcmp(parm, "Name"))
                                  {
                                    strncpy(
                                      eco_ptr->product[product_num]->name,
                                      val,
                                      ECO_PRODUCT_NAME_MAX
                                    );
                   eco_ptr->product[product_num]->name[ECO_PRODUCT_NAME_MAX - 1] = '\0';
                                  }
                                  /* SellPrice */
                                  else if(!strcmp(parm, "SellPrice"))
                                  {
                                    eco_ptr->product[product_num]->sell_price = atof(val);
                                  }
                                  /* BuyPrice */
                                  else if(!strcmp(parm, "BuyPrice"))
                                  {
                                    eco_ptr->product[product_num]->buy_price = atof(val);
                                  }
                                  /* Amount */
                                  else if(!strcmp(parm, "Amount"))
                                  {
                                    eco_ptr->product[product_num]->amount = atof(val);
                                  }
                                  /* AmountMax */
                                  else if(!strcmp(parm, "AmountMax"))
                                  {
                                    eco_ptr->product[product_num]->amount_max = atof(val);
                                  }
                                  /* EndEcoProduct */
                                  else if(!strcmp(parm, "EndEcoProduct"))
                                  {
                                    break;
                                  }

                                }
                            }	/* BeginEcoProduct */
 
		            /* EndEcoData */
		            else if(!strcmp(parm, "EndEcoData"))
		            {
/*
                          DBSortEconomyProducts(object_num);
 */
			        break;
		            }
		        }
		    }

                    /* ***************************************************** */
                    /* EndXSWObject */
                    else if(!strcmp(parm, "EndXSWObject"))
                    {
                        break;
                    }
                    /* Unsupported property */
		    else
		    {
		        fprintf(stderr,
		  "%s: Line %i: Warning: Unsupported parameter `%s'\n",
                            file, lines_read, parm
	                );
		    }
                }
	    }
	}


        /* Call notify progress function. */
        if(progress_notify != NULL)
            progress_notify(
                client_data,
                (int)filesize,
		(int)filesize
	    );
                                  

	/* Close file. */
	fclose(fp);
	fp = NULL;
                                  

	/* Make sure lost and found owner is valid. */
	if(header_buf != NULL)
	{
	    i = header_buf->lost_found_owner;

	    if((i < 0) || (i >= *total))
	        n = 1;
	    else if(ptr[i] == NULL)
		n = 1;
	    else if(ptr[i]->type != XSW_OBJ_TYPE_PLAYER)
		n = 1;
	    else
		n = 0;

	    if(n)
	    {
	        fprintf(stderr,
    "%s: Warning: Lost and found owner object #%i is invalid.\n",
		    file, i
	        );
	        header_buf->lost_found_owner = 0;
	    }
	}


	return(ptr);
}



#define WRITE	fwrite(buf, sizeof(char), strlen(buf), fp);

int UNVSaveToFile(
        char *file,
        xsw_object_struct **obj_ptr,
        int total,
        unv_head_struct *header_buf,
        void *client_data,
	void (*progress_notify)(void *, int, int)
)
{
	int i, n;
	char buf[1024];
	FILE *fp;

	xsw_ecodata_struct *eco_ptr;



	if(file == NULL)
	    return(-1);

	/* Open file for writing, overwrites existing one. */
	fp = fopen(file, "w");
	if(fp == NULL)
	{
            fprintf(stderr, "Unable to create file: %s\n",
                file
            );
            return(-1);
	}


	/* **************************************************** */
	/* Header */
        sprintf(buf, "# ShipWars Universe File\n");
        WRITE

        sprintf(buf, "#\n\n");
        WRITE

	if(header_buf != NULL)
	{
            sprintf(buf, "# Header\n");
            WRITE

            sprintf(buf, "BeginHeader\n");
            WRITE

            sprintf(buf, "    Title = %s\n", header_buf->title);
            WRITE

            sprintf(buf, "    Version = %s\n", header_buf->version);
            WRITE

            sprintf(buf, "    LostFoundOwner = %i\n", header_buf->lost_found_owner);
            WRITE

	    if(header_buf->ru_to_au < 1)
	        sprintf(buf, "    UnitConversionRUTOAU = %.2lf\n",
		    header_buf->ru_to_au);
	    else
	        sprintf(buf, "    UnitConversionRUTOAU = %.6lf\n",
		    header_buf->ru_to_au);
	    WRITE

            sprintf(buf, "    ImageSetFile = %s\n", header_buf->isr);
            WRITE
            sprintf(buf, "    OCSNFile = %s\n", header_buf->ocsn);
            WRITE
            sprintf(buf, "    SoundSchemeFile = %s\n", header_buf->ss);
            WRITE

            sprintf(buf, "EndHeader\n\n");
            WRITE
	}



	/* ********************************************************** */
        /* Begin writing objects to file. */

	for(i = 0; i < total; i++, obj_ptr++)
	{
            /* Call notify progress function. */   
            if(progress_notify != NULL)
                progress_notify(
                    client_data, 
                    i + 1,
                    total
                );

	    if(*obj_ptr == NULL)
		continue;

	    if((*obj_ptr)->type <= XSW_OBJ_TYPE_GARBAGE)
		continue;


	    /*   object_count is now of a valid object that needs
	     *   to be saved, begin saving xsw object number
	     *   object_count.
	     */

	    /* Comment. */
	    sprintf(buf, "# Object #%i\n", i);
            WRITE

	    /* BeginXSWObject */
	    sprintf(buf, "BeginXSWObject = %i\n", i);
	    WRITE

	    /* Name */
	    sprintf(buf, "    Name = %s\n", 
		(*obj_ptr)->name);
	    WRITE

            /* Password */
            sprintf(buf, "    Password = %s\n",
		(*obj_ptr)->password);  
            WRITE

	    /* Empire */
            sprintf(buf, "    Empire = %s\n",
		(*obj_ptr)->empire);
            WRITE

            /* Type */
            sprintf(buf, "    Type = %i\n",
                (*obj_ptr)->type);
            WRITE

            /* LocType */
            sprintf(buf, "    LocType = %i\n",
                (*obj_ptr)->loc_type);
            WRITE

            /* ImageSet */
            sprintf(buf, "    ImageSet = %i\n",
                (*obj_ptr)->imageset);
            WRITE

            /* Owner */
            sprintf(buf, "    Owner = %ld\n",
                (*obj_ptr)->owner);
            WRITE

            /* Size */
            sprintf(buf, "    Size = %ld\n",
                (*obj_ptr)->size);
            WRITE

            /* LockedOn */
            sprintf(buf, "    LockedOn = %ld\n",
                (*obj_ptr)->locked_on);
            WRITE

            /* InterceptingObject */
            sprintf(buf, "    InterceptingObject = %ld\n",
                (*obj_ptr)->intercepting_object);
            WRITE
 
            /* ScannerRange */
            sprintf(buf, "    ScannerRange = %.4lf\n",
                (*obj_ptr)->scanner_range);
            WRITE

            /* SectX */
            sprintf(buf, "    SectX = %ld\n",
                (*obj_ptr)->sect_x);
            WRITE
            /* SectY */
            sprintf(buf, "    SectY = %ld\n",
                (*obj_ptr)->sect_y);
            WRITE
            /* SectZ */
            sprintf(buf, "    SectZ = %ld\n",
                (*obj_ptr)->sect_z);
            WRITE

            /* CoordX */
            sprintf(buf, "    CoordX = %.4lf\n",
                (*obj_ptr)->x);
            WRITE
            /* CoordY */
            sprintf(buf, "    CoordY = %.4lf\n",
                (*obj_ptr)->y);
            WRITE
            /* CoordZ */
            sprintf(buf, "    CoordZ = %.4lf\n",
                (*obj_ptr)->z);
            WRITE

            /* ObjectHeading */
            sprintf(buf, "    ObjectHeading = %.4lf\n",
                (*obj_ptr)->heading);  
            WRITE

            /* ObjectPitch */
            sprintf(buf, "    ObjectPitch = %.4lf\n",
                (*obj_ptr)->pitch);
            WRITE

            /* ObjectBank */
            sprintf(buf, "    ObjectBank = %.4lf\n",
                (*obj_ptr)->bank);
            WRITE

            /* VelocityHeading */
            sprintf(buf, "    VelocityHeading = %.4lf\n",
                (*obj_ptr)->velocity_heading);  
            WRITE
                   
            /* VelocityMax */
            sprintf(buf, "    VelocityMax = %.4lf\n",
                (*obj_ptr)->velocity_max);
            WRITE
 
            /* Velocity */
            sprintf(buf, "    Velocity = %.4lf\n",
                (*obj_ptr)->velocity);    
            WRITE

            /* ThrustDir */
            sprintf(buf, "    ThrustDir = %.4lf\n",
                (*obj_ptr)->thrust_dir);
            WRITE

            /* ThrustPower */
            sprintf(buf, "    ThrustPower = %.4lf\n",
                (*obj_ptr)->thrust_power);
            WRITE

            /* Thrust */
            sprintf(buf, "    Thrust = %.4lf\n",
                (*obj_ptr)->thrust);
            WRITE

            /* Throttle */
            sprintf(buf, "    Throttle = %.4lf\n",
                (*obj_ptr)->throttle);      
            WRITE

            /* EngineState */
            sprintf(buf, "    EngineState = %i\n",
                (*obj_ptr)->engine_state);
            WRITE

            /* TurnRate */
            sprintf(buf, "    TurnRate = %.4lf\n",
                (*obj_ptr)->turnrate);
            WRITE

            /* Lighting */
            sprintf(buf, "    Lighting = %i\n",
                (*obj_ptr)->lighting);
            WRITE

            /* HitPointsMax */
            sprintf(buf, "    HitPointsMax = %.4lf\n",
                (*obj_ptr)->hp_max);
            WRITE

            /* HitPoints */
            sprintf(buf, "    HitPoints = %.4lf\n",
                (*obj_ptr)->hp);  
            WRITE
         
            /* PowerPurity */
            sprintf(buf, "    PowerPurity = %.4lf\n",
                (*obj_ptr)->power_purity);
            WRITE

            /* PowerMax */
            sprintf(buf, "    PowerMax = %.4lf\n",
                (*obj_ptr)->power_max);
            WRITE

            /* Power */
            sprintf(buf, "    Power = %.4lf\n",
                (*obj_ptr)->power);
            WRITE

            /* CoreEfficency */
            sprintf(buf, "    CoreEfficency = %.4lf\n",
                (*obj_ptr)->core_efficency);    
            WRITE

            /* AntimatterMax */
            sprintf(buf, "    AntimatterMax = %.4lf\n",
                (*obj_ptr)->antimatter_max);
            WRITE

            /* Antimatter */
            sprintf(buf, "    Antimatter = %.4lf\n",
                (*obj_ptr)->antimatter);
            WRITE

            /* ShieldState */
            sprintf(buf, "    ShieldState = %i\n",
                (*obj_ptr)->shield_state);
            WRITE

            /* ShieldFrequency */
            sprintf(buf, "    ShieldFrequency = %.4lf\n",
                (*obj_ptr)->shield_frequency);
            WRITE

            /* SelectedWeapon */
            sprintf(buf, "    SelectedWeapon = %i\n",
                (*obj_ptr)->selected_weapon);
            WRITE

            /* TotalWeapons */ 
	    if((*obj_ptr)->total_weapons > MAX_WEAPONS)
		(*obj_ptr)->total_weapons = MAX_WEAPONS;
            sprintf(buf, "    TotalWeapons = %i\n",
                (*obj_ptr)->total_weapons);
            WRITE

            /* BirthTime */
            sprintf(buf, "    BirthTime = %ld\n",
                (*obj_ptr)->birth_time);
            WRITE

            /* LifeSpan */
            sprintf(buf, "    LifeSpan = %ld\n",
                (*obj_ptr)->lifespan);
            WRITE

            /* CloakState */
            sprintf(buf, "    CloakState = %i\n",
                (*obj_ptr)->cloak_state);
            WRITE

            /* CloakStrength */
            sprintf(buf, "    CloakStrength = %.4lf\n",
                (*obj_ptr)->cloak_strength);
            WRITE

            /* ShieldVisibility */
            sprintf(buf, "    ShieldVisibility = %.4lf\n",
                (*obj_ptr)->shield_visibility);
            WRITE

	    /* CurrentVisibility */
            sprintf(buf, "    CurrentVisibility = %.4lf\n",  
                (*obj_ptr)->cur_visibility);
            WRITE

            /* Visibility */
            sprintf(buf, "    Visibility = %.4lf\n",
                (*obj_ptr)->visibility);
            WRITE

            /* DamageControl */
            sprintf(buf, "    DamageControl = %i\n",
                (*obj_ptr)->damage_control);
            WRITE

            /* ComChannel */
            sprintf(buf, "    ComChannel = %i\n",
                (*obj_ptr)->com_channel);
            WRITE


	    /* Scores. */
	    if((*obj_ptr)->score != NULL)
	    {
	        /* ScoreCredits */
                sprintf(buf, "    ScoreCredits = %.4lf\n",
                    (*obj_ptr)->score->credits);
                WRITE

                /* ScoreRMUMax */
                sprintf(buf, "    ScoreRMUMax = %.4lf\n",
		    (*obj_ptr)->score->rmu_max);
                WRITE

                /* ScoreRMU */
                sprintf(buf, "    ScoreRMU = %.4lf\n",
                    (*obj_ptr)->score->rmu); 
                WRITE
  
                /* ScoreDamageGivin */
                sprintf(buf, "    ScoreDamageGivin = %.4lf\n",
                    (*obj_ptr)->score->damage_givin);    
                WRITE

                /* ScoreDamageRecieved */
                sprintf(buf, "    ScoreDamageRecieved = %.4lf\n",
                    (*obj_ptr)->score->damage_recieved); 
                WRITE

                /* ScoreKills */
                sprintf(buf, "    ScoreKills = %ld\n",
                    (*obj_ptr)->score->kills); 
                WRITE
	    }

	    /* Permission stuff. */
            /* PermissionUID */
            sprintf(buf, "    PermissionUID = %i\n",
                (*obj_ptr)->permission.uid);
            WRITE

            /* PermissionGID */
            sprintf(buf, "    PermissionGID = %i\n",
                (*obj_ptr)->permission.gid); 
            WRITE

            /* Animation stuff. */
	    /* AnimationLastInterval */
            sprintf(buf, "    AnimationLastInterval = %ld\n",
                (*obj_ptr)->animation.last_interval);
            WRITE

            /* AnimationInterval */
            sprintf(buf, "    AnimationInterval = %ld\n",
                (*obj_ptr)->animation.interval);
            WRITE
 
            /* AnimationCurrentFrame */
            sprintf(buf, "    AnimationCurrentFrame = %i\n",
                (*obj_ptr)->animation.current_frame);
            WRITE

            /* AnimationTotalFrames */
            sprintf(buf, "    AnimationTotalFrames = %i\n",
                (*obj_ptr)->animation.total_frames);
            WRITE

            /* AnimationCycleCount */
            sprintf(buf, "    AnimationCycleCount = %i\n",
                (*obj_ptr)->animation.cycle_count);
            WRITE

            /* AnimationCycleTimes */
            sprintf(buf, "    AnimationCycleTimes = %i\n",
                (*obj_ptr)->animation.cycle_times);
            WRITE


	    /* Weapons. */
	    for(n = 0; n < (*obj_ptr)->total_weapons; n++)
	    {
		if((*obj_ptr)->weapons[n] == NULL)
		    continue;

		/* BeginWeapon */
		sprintf(buf, "    BeginWeapon\n");
		WRITE

                /* EmissionType */
                sprintf(buf, "        EmissionType = %i\n",
                (*obj_ptr)->weapons[n]->emission_type);
                WRITE

                /* OCSCode */
                sprintf(buf, "        OCSCode = %i\n",
                (*obj_ptr)->weapons[n]->ocs_code);
                WRITE

                /* Amount */
                sprintf(buf, "        Amount = %ld\n",
                (*obj_ptr)->weapons[n]->amount);
                WRITE
            
                /* Max */
                sprintf(buf, "        Max = %ld\n",
                (*obj_ptr)->weapons[n]->max);
                WRITE

		/* CreatePower */
                sprintf(buf, "        CreatePower = %.4lf\n",
                (*obj_ptr)->weapons[n]->create_power);
                WRITE

                /* Power */
                sprintf(buf, "        Power = %.4lf\n",
                (*obj_ptr)->weapons[n]->power);
                WRITE

                /* Range */
                sprintf(buf, "        Range = %ld\n",
                (*obj_ptr)->weapons[n]->range);
                WRITE
           
                /* Delay */
                sprintf(buf, "        Delay = %ld\n",
                (*obj_ptr)->weapons[n]->delay);
                WRITE

                /* LastUsed (always saved as 0) */
                sprintf(buf, "        LastUsed = 0\n");
                WRITE

                /* EndWeapon */
                sprintf(buf, "    EndWeapon\n");
                WRITE
	    }

	    /* Economy. */
	    eco_ptr = (*obj_ptr)->eco;
	    if(eco_ptr != NULL)
	    {
                sprintf(buf, "    BeginEcoData\n");
                WRITE

                sprintf(buf, "        Flags = %ld\n",
		    eco_ptr->flags
		);
                WRITE

                sprintf(buf, "        TaxGeneral = %.4lf\n",
                    eco_ptr->tax_general
                );
                WRITE

                sprintf(buf, "        TaxFriend = %.4lf\n",
                    eco_ptr->tax_friend
                );
                WRITE

                sprintf(buf, "        TaxHostile = %.4lf\n",
                    eco_ptr->tax_hostile
                );
                WRITE

		/* Products. */
		for(n = 0; n < eco_ptr->total_products; n++)
		{
		    if(eco_ptr->product[n] == NULL)
			continue;

                    sprintf(buf, "        BeginEcoProduct\n");
                    WRITE

                    sprintf(buf, "            Name = %s\n",
			eco_ptr->product[n]->name
		    );
                    WRITE

                    sprintf(buf, "            SellPrice = %.4lf\n", 
                        eco_ptr->product[n]->sell_price
                    );
                    WRITE

                    sprintf(buf, "            BuyPrice = %.4lf\n",
                        eco_ptr->product[n]->buy_price
                    );
                    WRITE

                    sprintf(buf, "            Amount = %.4lf\n",
                        eco_ptr->product[n]->amount
                    );
                    WRITE

                    sprintf(buf, "            AmountMax = %.4lf\n",
                        eco_ptr->product[n]->amount_max
                    );
                    WRITE

                    sprintf(buf, "        EndEcoProduct\n");
                    WRITE
		}


                sprintf(buf, "    EndEcoData\n");
                WRITE
	    }


            /* EndXSWObject */
            sprintf(buf, "EndXSWObject\n\n");
            WRITE
	}


        fclose(fp);   
	fp = NULL;


	return(0);
}
