/*
                             Widget: File IO


	Functions:

	image_t *WidgetLoadImageFromTgaFile(char *filename)
	image_t *WidgetLoadImageFromTgaData(u_int8_t *data)

	pixmap_t WidgetLoadPixmapFromTgaFile(char *filename)
        pixmap_t WidgetLoadPixmapFromTgaData(u_int8_t *data)


        image_t *WidgetLoadImageFromXpmFile(char *filename)
	image_t *WidgetLoadImageFromXpmData(char **data)

	pixmap_t WidgetLoadPixmapFromXpmFile(char *filename)
        pixmap_t WidgetLoadPixmapFromXpmData(char **data)

	---


*/


#include "../include/widget.h"

#include "../include/tga.h"     /* For tga loading in wfile.c */
#include <X11/xpm.h>


#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


/*
 *      Loads an image from a TGA file.
 *      The image's depth will be that of the GUI's.
 */
image_t *WidgetLoadImageFromTgaFile(char *filename)
{
        tga_data_struct td;
        image_t *image = NULL;
	void *ptr = NULL;
        off_t img_data_len = 0;
        int status = 0;


	/* Error checks. */
	if(!IDC())
	    return(NULL);


        /* Check if givin filename is valid. */
        if(filename == NULL)
        {
            fprintf(stderr,
      "WidgetLoadImageFromTgaFile(): Cannot load file with NULL name.\n"
            );
            return(NULL);     
        }


	/* Reset tga data struct. */
	bzero(&td, sizeof(tga_data_struct));


        /* Load tga file. */
        status = TgaReadFromFile(filename, &td, osw_gui[0].depth);
        switch(status)
        {
          case TgaNoBuffers:
            fprintf(stderr,
                "WidgetLoadImageFromTgaFile(): Error: Out of memory.\n"
            );
            break;

          case TgaBadHeader:
            fprintf(stderr,
                "WidgetLoadImageFromTgaFile(): %s: Error: Bad header.\n",
                filename
            );
            break;

          case TgaBadValue:
            fprintf(stderr,
                "WidgetLoadImageFromTgaFile(): %s: Error: Bad value.\n",
                filename
            );
            break;
              
          case TgaNoFile:
            fprintf(stderr,
                "WidgetLoadImageFromTgaFile(): %s: No such file.\n",
                filename
            );
            break;
              
          case TgaNoAccess:
            fprintf(stderr,
                "WidgetLoadImageFromTgaFile(): %s: Cannot read.\n",
                filename
            );
            break;

          default:
            break;
        }

	/* Return NULL on error. */
        if(status != TgaSuccess)
        {     
            TgaDestroyData(&td);
            return(NULL);  
        }

        /* Make sure SOURCE image depth is valid. */
        if((td.depth != 24) &&
           (td.depth != 32)
        )
        {
            fprintf(stderr,
           "WidgetLoadImageFromTgaFile(): %s: Incompatable bit depth %i.\n",
                filename,
                td.depth
            );
            TgaDestroyData(&td);
            return(NULL);
        }


	/* ************************************************************** */

        /*   Copy data in td to GUI image.
	 *
         *   Targa library loads data to memory in ZPixmap format
         *   of the requested depth.
	 */

	/* 8 bits. */
	switch(osw_gui[0].depth) {
	case 8:
	  if(OSWCreateImage(&image, td.width, td.height))
            {
	      fprintf(stderr,
		      "%s: Error: Cannot allocate GUI image.\n",
		      filename
		      );
            }
	  else
            {
	      img_data_len = MIN(
				 image->width * image->height * BYTES_PER_PIXEL8,
				 td.width * td.height * BYTES_PER_PIXEL8
				 );

	      ptr = memcpy(
			   image->data,        /* Target. */
			   td.data,            /* Source. */
			   img_data_len
			   );
            }
	  break;
	case 15:
	case 16:
        /* 15 or 16 bit. */
            if(OSWCreateImage(&image, td.width, td.height))
            {
                fprintf(stderr,
                    "%s: Error: Cannot allocate GUI image.\n",
                    filename   
                );
            }
            else
            {
                img_data_len = MIN(
                    image->width * image->height * BYTES_PER_PIXEL16,
                    td.width * td.height * BYTES_PER_PIXEL16
                );

                ptr = memcpy(
                    image->data,        /* Target. */
                    td.data,            /* Source. */
                    img_data_len
                );
            }
	    break;
        /* 24 or 32 bits. */
	case 24:
	case 32:
	    if(OSWCreateImage(&image, td.width, td.height))
	    {
                fprintf(stderr,
                    "%s: Error: Cannot allocate GUI image.\n",
                    filename
                );
            } 
	    else
	    {
                img_data_len = MIN(
                    image->width * image->height * BYTES_PER_PIXEL32,
		    td.width * td.height * BYTES_PER_PIXEL32
		);

                ptr = memcpy(
                    image->data,	/* Target. */
                    td.data,		/* Source. */
                    img_data_len
                );
	    }
        }


        /* Destroy tga data. */
        TgaDestroyData(&td);


        return(image);
}


/*
 *	Loads an image from TGA data in memory.
 *	The image's depth will be that of the GUI's.
 */
image_t *WidgetLoadImageFromTgaData(u_int8_t *data)
{
        tga_data_struct td;
	void *ptr;
        image_t *image;
        off_t img_data_len = 0;
        int status = 0;


	/* Error checks. */
	if(!IDC())
	    return(NULL);


        /* Check if givin data pointer is not NULL. */
        if(data == NULL)
        {
            fprintf(stderr,
      "WidgetLoadImageFromTgaData(): Cannot load file with NULL name.\n"
            );
            return(NULL);     
        }


	/* Reset tga data struct. */
	bzero(&td, sizeof(tga_data_struct));
	image = NULL;

        /* Load tga file. */
        status = TgaReadFromData(data, &td, osw_gui[0].depth);
        switch(status)
        {
          case TgaNoBuffers:
            fprintf(stderr,
                "WidgetLoadImageFromTgaData(): Error: Out of memory.\n"
            );
            break;

          case TgaBadHeader:
            fprintf(stderr,
                "WidgetLoadImageFromTgaData(): 0x%.8x: Error: Bad header.\n",
                (int)data
            );
            break;

          case TgaBadValue:
            fprintf(stderr,
                "WidgetLoadImageFromTgaData(): 0x%.8x: Error: Bad value.\n",
                (int)data
            );
            break;
              
          case TgaNoFile:
            fprintf(stderr,
                "WidgetLoadImageFromTgaData(): 0x%.8x: No such file.\n",
                (int)data
            );
            break;
              
          case TgaNoAccess:
            fprintf(stderr,
                "WidgetLoadImageFromTgaData(): 0x%.8x: Cannot read.\n",
                (int)data
            );
            break;

          default:
            break;
        }

	/* Return NULL on error. */
        if(status != TgaSuccess)
        {     
            TgaDestroyData(&td);
            return(NULL);  
        }

        /* Make sure SOURCE image data depth is acceptable. */
        if((td.depth != 24) &&
           (td.depth != 32)
        )
        {
            fprintf(stderr,
       "WidgetLoadImageFromTgaData(): 0x%.8x: Incompatable bit depth %i.\n",
                (int)data,
                (int)td.depth
            );
            TgaDestroyData(&td);
            return(NULL);
        }


	/* ************************************************************** */

        /*   Copy data in td to GUI image.
         *
         *   Targa library loads data to memory in ZPixmap format
	 *   of the requested depth.
	 */

	/* 8 bits. */
	switch(osw_gui[0].depth) {
	case 8:
            if(OSWCreateImage(&image, td.width, td.height))
            {
                fprintf(stderr,
                    "0x%.8x: Error: Cannot allocate GUI image.\n",
                    (u_int32_t)data
                );
            }
            else
            {   
                img_data_len = MIN(
                    image->width * image->height * BYTES_PER_PIXEL8,
                    td.width * td.height * BYTES_PER_PIXEL8
                );

                ptr = memcpy(
                    image->data,        /* Target. */
                    td.data,            /* Source. */
                    img_data_len
                );
            }
	    break;
	case 15:
	case 16:
        /* 16 bits. */
	  if(OSWCreateImage(&image, td.width, td.height))
            {
	      fprintf(stderr,
		      "0x%.8x: Error: Cannot allocate GUI image.\n",
		      (u_int32_t)data
		      );
            }
	  else
            {
	      img_data_len = MIN(image->width * image->height * BYTES_PER_PIXEL16,
				 td.width * td.height * BYTES_PER_PIXEL16);

	      ptr = memcpy(
			   image->data,        /* Target. */
			   td.data,            /* Source. */
			   img_data_len);
            }
	  break;
        /* 24 or 32 bits. */
	case 24:
	case 32:
	  if(OSWCreateImage(&image, td.width, td.height))
            {
	      fprintf(stderr,  
		      "0x%.8x: Error: Cannot allocate GUI image.\n",
		      (u_int32_t)data);
            }
	  else  
            {
	      img_data_len = MIN(
				 image->width * image->height * BYTES_PER_PIXEL32,
				 td.width * td.height * BYTES_PER_PIXEL32);

                ptr = memcpy(
                    image->data,        /* Target. */
                    td.data,            /* Source. */
                    img_data_len
                );
            }
        }


        /* Destroy tga data. */
        TgaDestroyData(&td);


        return(image);
}


/*
 *      Loads a pixmap from a TGA file.
 *      The pixmap's depth will match that of the GUI's.
 */
pixmap_t WidgetLoadPixmapFromTgaFile(char *filename)
{
        image_t *image = NULL;
        pixmap_t pixmap = 0;


        image = WidgetLoadImageFromTgaFile(filename);
        if(image == NULL)
            return(pixmap);

        pixmap = OSWCreatePixmapFromImage(image);
        OSWDestroyImage(&image);

        return(pixmap);
}


/*
 *      Loads a pixmap from TGA data in memory.
 *      The pixmap's depth will match that of the GUI's.
 */
pixmap_t WidgetLoadPixmapFromTgaData(u_int8_t *data)
{
	image_t *image = NULL;
	pixmap_t pixmap = 0;


	image = WidgetLoadImageFromTgaData(data);
	if(image == NULL)
	    return(pixmap);

	pixmap = OSWCreatePixmapFromImage(image);
	OSWDestroyImage(&image);

	return(pixmap);
}


/*
 *	Loads an image from an XPM file.
 *	The image's depth will match that of the GUI's.
 */
image_t *WidgetLoadImageFromXpmFile(char *filename)
{
        int status;
        image_t *image;
        image_t *imagemask;
        XpmAttributes xpmattr;


        /* Error checks. */
        if(!IDC() ||
           (osw_gui[0].root_win == 0) ||
           (filename == NULL)
        )
            return(0);


        /* ************************************************************ */
         
        /* Set XPM attributes. */
        bzero(&xpmattr, sizeof(XpmAttributes));
        xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
        xpmattr.closeness = XpmDefaultColorCloseness;
	xpmattr.depth = osw_gui[0].depth;

	/* Load image from XPM file. */
        status = XpmReadFileToImage(
            osw_gui[0].display,
            filename,
            &image,
            &imagemask,
            &xpmattr
        );
        if(status != XpmSuccess)
        {
            fprintf(stderr, "%s: Invalid Xpm file.\n", filename);
            return(0);
        }

        /* Destroy the mask image. */
	OSWDestroyImage(&imagemask);


	return(image);
}


/*
 *      Loads an image from XPM data in memory.
 *      The image's depth will match that of the GUI's.
 */
image_t *WidgetLoadImageFromXpmData(char **data)
{
        int status;
        image_t *image;
        image_t *imagemask;
        XpmAttributes xpmattr;


        /* Error checks. */
        if(!IDC() ||
           (osw_gui[0].root_win == 0) ||
           (data == NULL)
        )
            return(0);


        /* ************************************************************ */
            
        /* Set XPM attributes. */   
        bzero(&xpmattr, sizeof(XpmAttributes));
        xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
        xpmattr.closeness = XpmDefaultColorCloseness;
        xpmattr.depth = osw_gui[0].depth;

	/* Load image from XPM data. */
        status = XpmCreateImageFromData(
            osw_gui[0].display,
            data,
            &image,
            &imagemask,
            &xpmattr
        );
        if(status != XpmSuccess)
        {
            fprintf(stderr, "0x%.8x: Invalid Xpm data segment.\n",
		(u_int32_t)data
	    );
            return(0);
        }

	/* Destroy the mask image. */
        OSWDestroyImage(&imagemask);


        return(image);
}



/*
 *      Loads a pixmap from an XPM file.
 *      The pixmap's depth will match that of the GUI's.
 */
pixmap_t WidgetLoadPixmapFromXpmFile(char *filename)
{
	int status;
        pixmap_t pixmap = 0;
        pixmap_t pixmapmask = 0;
        XpmAttributes xpmattr;
	struct stat stat_buf;


	if(!IDC() ||
	   (osw_gui[0].root_win == 0) ||
	   (filename == NULL)
	)
	    return(0);


	/* Check if file exists. */
        if(stat(filename, &stat_buf))
        {
            fprintf(stderr,
                "%s: No such file.\n",
                filename
            );
            return(0);
        }

  
	/* ************************************************************ */

        /* Set XPM attributes. */
	bzero(&xpmattr, sizeof(XpmAttributes));
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
        xpmattr.closeness = XpmDefaultColorCloseness;
	xpmattr.depth = osw_gui[0].depth;

        /* Attempt to read the pixmap data from file. */
	status = XpmReadFileToPixmap(
	    osw_gui[0].display,
	    osw_gui[0].root_win,
	    filename,
	    &pixmap,
	    &pixmapmask,
	    &xpmattr
	);
	if(status != XpmSuccess)
        {
            fprintf(stderr, "%s: Invalid Xpm file.\n", filename);
            return(0);
        }


	/* Destroy the mask, we don't need it. */
	OSWDestroyPixmap(&pixmapmask);


	return(pixmap);
}


/*
 *      Loads a pixmap from XPM data in memory.
 *      The pixmap's depth will match that of the GUI's.
 */
pixmap_t WidgetLoadPixmapFromXpmData(char **data)  
{
        int status;  
        pixmap_t pixmap = 0;
        pixmap_t pixmapmask = 0;
        XpmAttributes xpmattr;
            
          
        if(!IDC() ||
           (osw_gui[0].root_win == 0) ||
           (data == NULL)
        )
            return(0);


        /* ************************************************************ */

        /* Set XPM attributes. */
        bzero(&xpmattr, sizeof(XpmAttributes));
        xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
        xpmattr.closeness = XpmDefaultColorCloseness;
	xpmattr.depth = osw_gui[0].depth;

        /* Attempt to read the pixmap data from file. */
        status = XpmCreatePixmapFromData(
            osw_gui[0].display,
            osw_gui[0].root_win,
            data,
            &pixmap,
            &pixmapmask,
            &xpmattr
        );
        if(status != XpmSuccess)
        {
            fprintf(stderr,
		"0x%.8x: Invalid Xpm data segment.\n",
		(u_int32_t)data
	    );
            return(0);
        }
 
        
        /* Destroy the mask, we don't need it. */
        OSWDestroyPixmap(&pixmapmask);


        return(pixmap);
}
