/*
                 Joystick IO and Joystick Configuration IO

	Functions:

	int JSOpen(
	        char *js_device,
	        jscal_rec_struct *jsd
	)
	int JSRead(jscal_rec_struct *jsd)
	void JSClose(jscal_rec_struct *jsd)

	int JSReadCaliberationFile(char *filename)
	int JSWriteCaliberationFile(char *filename)

	---
 */
   
#include "xjscal.h"


#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))


/*
 *	Opens the joystick js_device and setting the proper info
 *	into jsd.
 *
 *	jsd members exists and fd will be modified to contain the
 *	correct values.
 */
int JSOpen(
	char *js_device,
	jscal_rec_struct *jsd
)
{
	int i;


	if(jsd == NULL)
	    return(-1);

	/* Default to device name DEV_JOYSTICK_0 if NULL. */
	if(js_device == NULL)
	    js_device = DEV_JOYSTICK_0;

        /* Open the joystick device. */
        i = open(js_device, O_RDONLY);
	if(i < 0)
	{
	    jsd->exists = 0;
	    jsd->fd = -1;
	}
	else
	{
            jsd->exists = 1;
            jsd->fd = i;
	}


	return(i);
} 

/*
 *	Updates the button states and coordinate position members
 *	of jsd.
 *
 *	Returns non-zero on error.
 */
int JSRead(jscal_rec_struct *jsd)
{
	static int status;
	static struct JS_DATA_TYPE js_data_buf;


	if(jsd == NULL)
	    return(-1);

	if(jsd->fd < 0)
	    return(-1);


	/* Read joystick data. */
	status = read(jsd->fd, &js_data_buf, JS_RETURN);  


	/* Record button states. */
        jsd->prev_buttons = jsd->buttons;  
	/* Set new button states. */
        jsd->buttons = (u_int64_t)js_data_buf.buttons;

	/* Set coordiantes. */
	jsd->x = js_data_buf.x;
	jsd->y = js_data_buf.y;

	/* Calculate coeff coordinates. */
	if(jsd->x < jsd->x_center)
	{
	    jsd->x_coeff = (double)((double)(jsd->x - jsd->x_min) / 
                (double)MAX(jsd->x_center - jsd->x_min, 1)) - 1;
	}
	else
        {
            jsd->x_coeff = (double)(jsd->x - jsd->x_center) /
                (double)MAX(jsd->x_max - jsd->x_center, 1);
        }

        if(jsd->y < jsd->y_center)
        {
            jsd->y_coeff = (double)((double)(jsd->y - jsd->y_min) /
                (double)MAX(jsd->y_center - jsd->y_min, 1)) - 1;
        }   
        else
        {
            jsd->y_coeff = (double)(jsd->y - jsd->y_center) /
                (double)MAX(jsd->y_max - jsd->y_center, 1);
        }
 


	return(status);
}


/*
 *	Closes the joystick referanced in jsd.
 *
 *	If the joystick referanced is already closed, no operation
 *	will be performed.
 */
void JSClose(jscal_rec_struct *jsd)
{
        if(jsd == NULL)
            return;

	if(jsd->fd > -1)
	{
	    close(jsd->fd);
	    jsd->fd = -1;
	}

	return;
}


int JSReadCaliberationFile(char *filename)
{
        char *strptr, *strptr2;

        FILE *fp;
        off_t filesize;
        struct stat stat_buf;

        char parm[CFG_PARAMETER_MAX];
        char val[CFG_VALUE_MAX];
        int lines_read = 0; 

	int js_num;


        /* Check if filename exists. */
        if(filename == NULL)
            return(-1);
        if(stat(filename, &stat_buf))
        {
	    /* Do not print error. */
            return(-1);
        }

        /* Get size of file. */
        filesize = stat_buf.st_size;

        /* Open filename. */
        fp = fopen(filename, "r");
        if(fp == NULL)
        {
            fprintf(stderr, "%s: Cannot open.\n", filename);
            return(-1);
        }

        /* ************************************************************* */
	/* Reset joystick caliberation data. */
	for(js_num = 0; js_num < MAX_JOYSTICKS; js_num++)
	{
	    jscal_rec[js_num].exists = 0;

            jscal_rec[js_num].x_min = DEFAULT_X_AXIS_MIN_VALUE;
	    jscal_rec[js_num].x_center = ((DEFAULT_X_AXIS_MAX_VALUE -
		DEFAULT_X_AXIS_MIN_VALUE) / 2) +
		DEFAULT_X_AXIS_MIN_VALUE;
            jscal_rec[js_num].x_max = DEFAULT_X_AXIS_MAX_VALUE;

            jscal_rec[js_num].y_min = DEFAULT_Y_AXIS_MIN_VALUE;
            jscal_rec[js_num].y_center = ((DEFAULT_Y_AXIS_MAX_VALUE -
                DEFAULT_Y_AXIS_MIN_VALUE) / 2) +
                DEFAULT_Y_AXIS_MIN_VALUE;
            jscal_rec[js_num].y_max = DEFAULT_Y_AXIS_MAX_VALUE;

	    jscal_rec[js_num].x_scope = jscal_rec[js_num].x_max -
		jscal_rec[js_num].x_min;
            jscal_rec[js_num].y_scope = jscal_rec[js_num].y_max -
                jscal_rec[js_num].y_min;

            jscal_rec[js_num].nullzone = DEFAULT_NULLZONE_VALUE;

            jscal_rec[js_num].x_trim = 0;
            jscal_rec[js_num].y_trim = 0;
	}

        /* ******************************************************** */

        strptr = NULL;

        while(1)
        {
            /* Free previous line and allocate/read next line. */
            free(strptr); strptr = NULL;
            strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
            if(strptr == NULL) break;
            lines_read++;
         
            /* Fetch parameter. */
            strptr2 = StringCfgParseParm(strptr);
            if(strptr2 == NULL) continue;
            strncpy(parm, strptr2, CFG_PARAMETER_MAX);
            parm[CFG_PARAMETER_MAX - 1] = '\0';

            /* Fetch value. */
            strptr2 = StringCfgParseValue(strptr);
            if(strptr2 == NULL) strptr2 = "0";  /* Set it to "0" if NULL. */
            strncpy(val, strptr2, CFG_VALUE_MAX);
            val[CFG_VALUE_MAX - 1] = '\0';

	    /* BeginJoystick */
            if(strpfx(parm, "BeginJoystick"))	/* Use strfpx() */
            {
		if(!strcmp(parm, "BeginJoystick1"))
		    js_num = 0;
                else if(!strcmp(parm, "BeginJoystick2"))
                    js_num = 1;
                else if(!strcmp(parm, "BeginJoystick3"))
                    js_num = 2;
                else if(!strcmp(parm, "BeginJoystick4"))
                    js_num = 3;
                else
                    js_num = 0;		/* Assume 0 if not numbered. */


                while(1)
                {
                    /* Free previous line and allocate/read next line. */
                    free(strptr); strptr = NULL;
                    strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
                    if(strptr == NULL) break;
                    lines_read++;

                    /* Fetch parameter. */
                    strptr2 = StringCfgParseParm(strptr);
                    if(strptr2 == NULL) continue;
                    strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                    parm[CFG_PARAMETER_MAX - 1] = '\0';

                    /* Fetch value. */
                    strptr2 = StringCfgParseValue(strptr);
                    if(strptr2 == NULL) strptr2 = "0";  /* Set it to "0" if NULL. */
                    strncpy(val, strptr2, CFG_VALUE_MAX);
                    val[CFG_VALUE_MAX - 1] = '\0';


		    /* X_Min */
                    if(!strcmp(parm,"X_Min"))
                    {
                        jscal_rec[js_num].x_min = atoi(val);
                    }
                    /* X_Center */
                    else if(!strcmp(parm,"X_Center"))
                    {
                        jscal_rec[js_num].x_center = atoi(val);
                    }
                    /* X_Max */
                    else if(!strcmp(parm, "X_Max"))
                    {
                        jscal_rec[js_num].x_max = atoi(val);
                    }

                    /* Y_Min */
                    else if(!strcmp(parm, "Y_Min"))
                    {
                        jscal_rec[js_num].y_min = atoi(val);
                    }
                    /* Y_Center */
                    else if(!strcmp(parm,"Y_Center"))
                    {
                        jscal_rec[js_num].y_center = atoi(val);
                    }
                    /* Y_Max */
                    else if(!strcmp(parm, "Y_Max"))
                    {
                        jscal_rec[js_num].y_max = atoi(val);
                    }

                    /* NullZone */
                    else if(!strcmp(parm, "NullZone"))
                    {
                        jscal_rec[js_num].nullzone = atoi(val);
                    }

                    /* X_Trim */
                    else if(!strcmp(parm, "X_Trim"))
                    {
                        jscal_rec[js_num].x_trim = atoi(val);
                    }
                    /* Y_Trim */
                    else if(!strcmp(parm, "Y_Trim"))
                    {
                        jscal_rec[js_num].y_trim = atoi(val);
                    }  

                    /* EndJoystick */
                    else if(strpfx(parm, "EndJoystick")) /* Use strpfx() */
                    {
			break;
                    }
		}
	    }
	}

	/* ******************************************************* */

        /* Close file. */
        fclose(fp);
	fp = NULL;


	/* Calculate scope values. */
        for(js_num = 0; js_num < MAX_JOYSTICKS; js_num++)
        {
            jscal_rec[js_num].x_scope = jscal_rec[js_num].x_max -
                jscal_rec[js_num].x_min;

            jscal_rec[js_num].y_scope = jscal_rec[js_num].y_max -
                jscal_rec[js_num].y_min;
        }


	return(0);
}



int JSWriteCaliberationFile(char *filename)
{
        char stringa[256];
        FILE *fp;
	int js_num;

	int bytes_written = 0;


        /* Open dbfilename for writing, overwrites existing one. */
        fp = fopen(filename, "w");
        if(fp == NULL)
            return(-1);


        /* ****************************************************** */
        /* Write header. */
        sprintf(stringa,
"# Joystick Caliberation File\n#\n"
	);
        bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

        sprintf(stringa,
"#      Contains the caliberation information for your joysticks.\n#\n"
        );
        bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

        sprintf(stringa,
"#      Created by %s Version %s.\n#\n",
		XJSCAL_NAME, XJSCAL_VERSION
	);
        bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);


        /* ******************************************************** */
        /* ***   Begin Writing Joysticks Caliberation to File   *** */
	for(js_num = 0; js_num < MAX_JOYSTICKS; js_num++)
	{

           /* Comment */
           sprintf(stringa, "\n# Joystick %i.\n", (js_num + 1) );
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

           /* BeginJoystick# */
           sprintf(stringa, "BeginJoystick%i\n", (js_num + 1) );
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);


           /* X_Min */
           sprintf(stringa, "   X_Min = %i\n",
                jscal_rec[js_num].x_min);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

           /* X_Center */
           sprintf(stringa, "   X_Center = %i\n",
                jscal_rec[js_num].x_center);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

           /* X_Max */
           sprintf(stringa, "   X_Max = %i\n",
                jscal_rec[js_num].x_max);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);


           /* Y_Min */
           sprintf(stringa, "   Y_Min = %i\n",
                jscal_rec[js_num].y_min);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

           /* Y_Center */
           sprintf(stringa, "   Y_Center = %i\n",
                jscal_rec[js_num].y_center);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

           /* Y_Max */
           sprintf(stringa, "   Y_Max = %i\n",
                jscal_rec[js_num].y_max);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);


           /* NullZone */
           sprintf(stringa, "   NullZone = %i\n",
                jscal_rec[js_num].nullzone);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

           /* X_Trim */
           sprintf(stringa, "   X_Trim = %i\n",
                jscal_rec[js_num].x_trim);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

           /* Y_Trim */
           sprintf(stringa, "   Y_Trim = %i\n",
                jscal_rec[js_num].y_trim);
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);

           /* EndJoystick# */
           sprintf(stringa, "EndJoystick%i\n", (js_num + 1) );
           bytes_written += fwrite(stringa,sizeof(char),strlen(stringa),fp);
	}


	/* Close the file. */
        fclose(fp);

	return(0);
}
