/*
              X J O Y S T I C K   C A L I B R A T O R

Copyright (C) 1997, 1998, 1999 WolfPack Development Organization.

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.


                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
WOLFPACK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

WOLFPACK DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
WOLFPACK BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

 */


#ifndef XJSCAL_H
#define XJSCAL_H


#include <stdio.h>
#include <db.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
extern int errno;
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>

#include <sys/time.h>
#include <sys/stat.h>

#if defined(__linux__)
    #include <linux/unistd.h>
    #include <linux/limits.h>

    #include <linux/joystick.h>
#else

#endif


/* Local header files. */
#include "../include/os.h"	/* OS `safety net' definations. */
#include "../include/fio.h"
#include "../include/cfgfmt.h"	/* Standard configuration format. */
#include "../include/graphics.h"        /* Blitting stuff. */
#include "../include/strexp.h"
#include "../include/string.h"
#include "../include/disk.h"

#include "../include/tga.h"             /* Targa library. */

#include "../include/osw-x.h"
#include "../include/widget.h"          /* Widget system and OSW. */

#include "blitting.h"


/* ***************************************************************** */

/*
 *   Default Joystick Caliberation File:
 *
 *	This is the caliberation file to read and write joystick
 *	caliberation information to.   An alternative name for
 *	this file can be specified at the command line.
 */
#define JS_CALIBRATION_FILE	".joystick"

/*
 *   Joystick Device Nodes:
 *
 *	These are the device node names for your joysticks. Currently
 *	the Linux joystick driver only has support for joystick 1
 *	(/dev/js0) and joystick 2 (/dev/js1).
 *
 */
#define DEV_JOYSTICK_0	"/dev/js0"
#define DEV_JOYSTICK_1	"/dev/js1"
#define DEV_JOYSTICK_2	"/dev/js2"
#define DEV_JOYSTICK_3	"/dev/js3"

/*
 *   Maximum Number of Joysticks:
 *
 *	This is the maximum number of joysticks that this program
 *	will support.
 */
#define MAX_JOYSTICKS	4



/*
 *   Default Minimum and Maximum X Y Values:
 *
 *	These are the values to be assumed when no values are available
 *	for a joystick's x and y axis minimum and maximum values.
 */
#define DEFAULT_X_AXIS_MIN_VALUE	0
#define DEFAULT_X_AXIS_MAX_VALUE	1200
#define DEFAULT_Y_AXIS_MIN_VALUE	0
#define DEFAULT_Y_AXIS_MAX_VALUE	1200
#define DEFAULT_NULLZONE_VALUE		120


/*
 *   Maximum X and Y Joystick Values:
 *
 *	Some joysticks have a very large range of X and Y axis values.
 *	Normally, joysticks have X and Y axis values from 0 to 1400.
 *	But some joysticks may produce values outside of this range.
 *	Some kind of limit must be set by this program in order to limit
 *	and truncate huge numbers.
 */
#define MIN_JS_X_AXIS_VALUE	-16384
#define MAX_JS_X_AXIS_VALUE	16383
#define MIN_JS_Y_AXIS_VALUE	-16384
#define MAX_JS_Y_AXIS_VALUE	16383


/*
 *   XJoystick Caliberator Loop Interval:
 *
 *	This controls the interval for the main loop of this program.
 *	Units are in microseconds.
 */
#define PROG_LOOP_INTERVAL	16000



/*
 *   Master name and version strings:
 */
#define XJSCAL_NAME	"XJoystick Caliberator"
#define XJSCAL_VERSION	"2.1"



/*
 *    Usage message:
 */
#define XJSCAL_USAGE "\
Usage: xjscal [options] [GUI_options]\n\
\n\
    [options] can be any of the following:\n\
\n\
        --device <device>      Specify startup joystick device\n\
                               (default is /dev/js0).\n\
        --file <path>          Specify caliberation file\n\
                               (default is $HOME/.joystick).\n\
        --help                 Prints (this) help screen and exits.\n\
        --version              Prints version information and exits.\n\
\n\
    [GUI_options] can be any options standard to your GUI, consult\n\
    your GUI's manual for available options.\n\
\n"


/*
 *   Color Specifications:
 */
#define NORMAL_TEXT_COLOR_SPEC      "rgbi:1.00/1.00/1.00"
#define BOLD_TEXT_COLOR_SPEC        "rgbi:1.00/1.00/1.00"
#define READOUT_TEXT_COLOR_SPEC     "rgbi:1.00/0.18/0.10"



/* ********************************************************************
 *
 *                       GLOBAL VARIABLES
 *
 */

/*
 *   Process handling:
 *
 *      Note the rootpid is the PID of the main xpppcon process,
 *      the process that started it all.   childpid can change
 *      during execution, it will contain and child processes'
 *      PID.
 */
pid_t rootpid;
pid_t childpid;
int child_procs;


int runlevel;


/*
 *	Current time since midnight in milliseconds:
 */
long cur_millitime;


/*
 *      File names:   
 */
struct {

	/* Joystick caliberation file. */
	char js_calib[PATH_MAX + NAME_MAX];              

	/* Device name of current selected joystick. */
        char js_device[PATH_MAX + NAME_MAX];

} fname;


/*
 *	Timming scheduals:
 *
 *	In milliseconds unless otherwise noted.
 */
struct {

	time_t	redraw_buttons,
		redraw_coordinates;

} next;


/*
 *	Joystick data record structure:
 */
typedef struct
{
	int exists;		/* 1 for exists, 0 for non-existant. */
	int fd;			/* Descriptor, -1 for not opened. */

	/* Current values. */
	int x, y;		/* Current x and y values. */
	double x_coeff, y_coeff;	/* -1.0 to 1.0 */

	u_int64_t prev_buttons;	/* Previous buttons state mask. */
	u_int64_t buttons;	/* Buttons state mask. */


	/* Caliberation info. */
	int x_min, x_center, x_max;
	int y_min, y_center, y_max;
	int x_scope, y_scope;		/* Max - min. */

	unsigned int nullzone;		/* Null zone radius. */


	/* Trim. */
	int x_trim, y_trim;

} jscal_rec_struct;
jscal_rec_struct jscal_rec[MAX_JOYSTICKS];



/*
 *	Pixels:
 */
pixel_t		normal_text_pixel,
		bold_text_pixel,
		readout_text_pixel;

/*
 *	Joystick icon pixmap:
 */
pixmap_t	js_icon_pm;


/*
 *	X Joystick Caliberator Window Structure:
 */
#define MODE_SWITCH_TESTING	0
#define MODE_SWITCH_CALIB	1

struct {

	bool_t map_state;
	bool_t focused;

	win_t	toplevel,
		coord_readout,
		btn_readout,		
		mode_switch,
		js_num_switch,
		save_btn,
		exit_btn;

	bool_t	save_btn_state,
		exit_btn_state;

	pixmap_t	toplevel_pm,
			exit_btn0_pm,
			exit_btn1_pm,
			mode_switch0_pm,
			mode_switch1_pm,
			num_switch0_pm,
			num_switch1_pm,
			num_switch2_pm,
			num_switch3_pm,
			readout_pm,	/* Button readouts. */
			tmp_buf_pm,	/* Buffer for readout pixmap. */
			save_btn0_pm,
			save_btn1_pm;

	image_t *coord_readout_img;
	shared_image_t *coord_readout_buf_img;
	image_t *crosshair_img;

	/* Text message images. */
	image_t *calib_text_img,
		*nojs_text_img;

	/* Selected joystick number (for js_num_switch). */
	int sel_js_num;		/* 0 to MAX_JOYSTICKS - 1 */

	/* Mode switch state (for mode_switch). */
	int mode_switch_state;	/* One of MODE_SWITCH_ * */


} xjscal_win;

dialog_win_struct dialog;
fbrowser_struct fb;



/* *************************************************************** */

/* In joystick.c */
extern int JSOpen(
	char *device,
	jscal_rec_struct *jsd
);
extern int JSRead(jscal_rec_struct *jsd);
extern void JSClose(jscal_rec_struct *jsd);

extern int JSReadCaliberationFile(char *filename);
extern int JSWriteCaliberationFile(char *filename);


/* In timming.c */
extern long MilliTime(void);

/* In xjscal.c */
extern void XJSCalHelp();
extern void XJSCalVersion();

extern int XJSCalFBOK(char *path);
extern int XJSCalFBCancel(char *path);

extern void XJSCalUnfocusAll();
extern void XJSCalShutdown();

extern void XJSCalDrawButtons();
extern void XJSCalDrawSwitches();
extern void XJSCalDrawReadout();
extern void XJSCalDrawCoords();
extern void XJSCalDrawAll();

extern void XJSCalDoCalib();

extern void XJSCalResetTimmers();


#endif /* XJSCAL_H */
